# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from .utils import WorkerActor

logger = logging.getLogger(__name__)


class ProcessHelperActor(WorkerActor):
    """
    Actor handling utils on every process
    """
    def __init__(self):
        super().__init__()
        self._dispatch_ref = None
        self._daemon_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor
        from .daemon import WorkerDaemonActor

        super().post_create()
        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())
        self._dispatch_ref.register_free_slot(self.uid, 'process_helper')

        self._daemon_ref = self.ctx.actor_ref(WorkerDaemonActor.default_uid())
        if self.ctx.has_actor(self._daemon_ref):
            self._daemon_ref.register_process(self.ref(), os.getpid(), _tell=True)
        else:
            self._daemon_ref = None

    @staticmethod
    def free_mkl_buffers():
        """
        Free MKL buffer
        """
        from ..lib.mkl_interface import mkl_free_buffers
        if mkl_free_buffers is None:
            return
        mkl_free_buffers()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...serialize import IndexField, DataTypeField, KeyField, BoolField
from ...tensor.utils import get_chunk_slices
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, decide_series_chunk_size


class IndexDataSource(DataFrameOperand, DataFrameOperandMixin):
    """
    Represent data from pandas Index
    """

    _op_type_ = OperandDef.INDEX_DATA_SOURCE

    _input = KeyField('input')
    _data = IndexField('data')
    _dtype = DataTypeField('dtype')
    _store_data = BoolField('store_data')

    def __init__(self, input=None, data=None, dtype=None, gpu=None,  # pylint: disable=redefined-builtin
                 store_data=None, sparse=None, **kw):
        super().__init__(_input=input, _data=data, _dtype=dtype, _gpu=gpu, _sparse=sparse,
                         _store_data=store_data, _output_types=[OutputType.index], **kw)

    @property
    def input(self):
        return self._input

    @property
    def data(self):
        return self._data

    @property
    def dtype(self):
        return self._dtype

    @property
    def store_data(self):
        return self._store_data

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if inputs is not None and len(inputs) > 0:
            self._input = self._inputs[0]

    def __call__(self, shape=None, chunk_size=None, inp=None, name=None,
                 names=None):
        if inp is None:
            # create from pandas Index
            name = name if name is not None else self._data.name
            names = names if names is not None else self._data.names
            return self.new_index(None, shape=shape, dtype=self._dtype,
                                  index_value=parse_index(self._data, store_data=self.store_data),
                                  name=name, names=names, raw_chunk_size=chunk_size)
        elif hasattr(inp, 'index_value'):
            # get index from Mars DataFrame, Series or Index
            name = name if name is not None else inp.index_value.name
            names = names if names is not None else [name]
            if inp.index_value.has_value():
                self._data = data = inp.index_value.to_pandas()
                return self.new_index(None, shape=(inp.shape[0],), dtype=data.dtype,
                                      index_value=parse_index(data, store_data=self.store_data),
                                      name=name, names=names, raw_chunk_size=chunk_size)
            else:
                if self._dtype is None:
                    self._dtype = inp.index_value.to_pandas().dtype
                return self.new_index([inp], shape=(inp.shape[0],),
                                      dtype=self._dtype, index_value=inp.index_value,
                                      name=name, names=names)
        else:
            if inp.ndim != 1:
                raise ValueError('Index data must be 1-dimensional')
            # get index from tensor
            dtype = inp.dtype if self._dtype is None else self._dtype
            pd_index = pd.Index([], dtype=dtype)
            if self._dtype is None:
                self._dtype = pd_index.dtype
            return self.new_index([inp], shape=inp.shape, dtype=self._dtype,
                                  index_value=parse_index(pd_index, inp, store_data=self.store_data),
                                  name=name, names=names)

    @classmethod
    def _tile_from_pandas(cls, op):
        index = op.outputs[0]
        raw_index = op.data

        memory_usage = raw_index.memory_usage(deep=True)
        chunk_size = index.extra_params.raw_chunk_size or options.chunk_size
        chunk_size = decide_series_chunk_size(index.shape, chunk_size, memory_usage)
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_index, chunk_shape in zip(itertools.product(*chunk_size_idxes),
                                            itertools.product(*chunk_size)):
            chunk_op = op.copy().reset_key()
            slc = get_chunk_slices(chunk_size, chunk_index)
            chunk_op._data = chunk_data = raw_index[slc]
            out_chunk = chunk_op.new_chunk(
                None, shape=chunk_shape, dtype=index.dtype, index=chunk_index,
                name=index.name, index_value=parse_index(chunk_data, store_data=op.store_data))
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_indexes(None, index.shape, dtype=index.dtype,
                                  index_value=index.index_value,
                                  name=index.name, chunks=out_chunks, nsplits=chunk_size)

    @classmethod
    def _tile_from_dataframe(cls, op):
        inp = op.input
        out = op.outputs[0]

        out_chunks = []
        if inp.ndim == 1:
            # series, index
            for c in inp.chunks:
                chunk_op = op.copy().reset_key()
                out_chunk = chunk_op.new_chunk([c], shape=c.shape,
                                               dtype=out.dtype, index=c.index,
                                               index_value=c.index_value,
                                               name=out.name)
                out_chunks.append(out_chunk)
            nsplits = inp.nsplits
        else:
            # DataFrame
            nsplit = inp.nsplits[1]
            axis_1_index = np.argmin(nsplit).item()
            for i in range(inp.chunk_shape[0]):
                chunk_index = (i, axis_1_index)
                c = inp.cix[chunk_index]
                chunk_op = op.copy().reset_key()
                out_chunk = chunk_op.new_chunk([c], shape=(c.shape[0],),
                                               dtype=out.dtype, index=(i,),
                                               index_value=c.index_value,
                                               name=out.name)
                out_chunks.append(out_chunk)
            nsplits = (inp.nsplits[0],)

        new_op = op.copy()
        params = out.params
        params['chunks'] = out_chunks
        params['nsplits'] = nsplits
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile_from_tensor(cls, op):
        inp = op.input
        out = op.outputs[0]
        out_chunks = []
        for c in inp.chunks:
            chunk_op = op.copy().reset_key()
            index_value = parse_index(out.index_value.to_pandas(), c, store_data=op.store_data)
            out_chunk = chunk_op.new_chunk([c], shape=c.shape,
                                           dtype=out.dtype, index=c.index,
                                           index_value=index_value,
                                           name=out.name)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out.params
        params['chunks'] = out_chunks
        params['nsplits'] = inp.nsplits
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def tile(cls, op):
        if op.input is None:
            # from pandas
            return cls._tile_from_pandas(op)
        elif hasattr(op.input, 'index_value'):
            # from DataFrame or Series
            return cls._tile_from_dataframe(op)
        else:
            # from tensor
            return cls._tile_from_tensor(op)

    @classmethod
    def execute(cls, ctx, op):
        if op.input is None:
            # from pandas
            ctx[op.outputs[0].key] = op.data
        else:
            out = op.outputs[0]
            inp = ctx[op.input.key]
            dtype = out.dtype if out.dtype != np.object else None
            if hasattr(inp, 'index'):
                # DataFrame, Series
                ctx[out.key] = pd.Index(inp.index, dtype=dtype, name=out.name)
            else:
                ctx[out.key] = pd.Index(inp, dtype=dtype, name=out.name)


def from_pandas(data, chunk_size=None, gpu=False, sparse=False, store_data=False):
    op = IndexDataSource(data=data, gpu=gpu, sparse=sparse, dtype=data.dtype,
                         store_data=store_data)
    return op(shape=data.shape, chunk_size=chunk_size)


def from_tileable(tileable, dtype=None, name=None, names=None):
    op = IndexDataSource(gpu=tileable.op.gpu, sparse=tileable.issparse(),
                         dtype=dtype)
    return op(inp=tileable, name=name, names=names)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import time
import logging
import pickle
import sys
import uuid
from io import BytesIO
from numbers import Integral

import numpy as np

from ..config import options
from ..errors import ResponseMalformed, ExecutionInterrupted, ExecutionFailed, \
    ExecutionStateUnknown, ExecutionNotStopped
from ..operands import Fetch
from ..serialize import dataserializer
from ..serialize.dataserializer import pyarrow
from ..tensor.core import Indexes
from ..utils import build_tileable_graph, sort_dataframe_result, numpy_dtype_from_descr_json

logger = logging.getLogger(__name__)


class Session(object):
    def __init__(self, endpoint, session_id=None, req_session=None, verify_ssl=True,
                 **session_kwargs):
        self._endpoint = endpoint.rstrip('/')
        self._session_id = session_id
        self._session_kwargs = session_kwargs
        # dict structure: {tileable_key -> graph_key, tileable_ids}
        # dict value is a tuple object which records graph key and tileable id
        self._executed_tileables = dict()

        self._serial_type = None
        self._pickle_protocol = pickle.HIGHEST_PROTOCOL

        if req_session:
            self._req_session = req_session
        else:
            import requests
            from requests.adapters import HTTPAdapter

            self._req_session = requests.Session()
            self._req_session.mount('http://stackoverflow.com', HTTPAdapter(max_retries=5))

        self._req_session.verify = verify_ssl
        if not verify_ssl:
            try:
                import urllib3
                urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            except ImportError:  # pragma: no cover
                pass

        self._main()

    @property
    def session_id(self):
        return self._session_id

    @property
    def endpoint(self):
        return self._endpoint

    @endpoint.setter
    def endpoint(self, url):
        self._endpoint = url

    def _main(self):
        if pyarrow is None:
            self._serial_type = dataserializer.SerialType.PICKLE
        else:
            self._serial_type = dataserializer.SerialType(options.client.serial_type.lower())

        session_kw = self._session_kwargs.copy()
        session_kw['pyver'] = '.'.join(str(v) for v in sys.version_info[:3])
        session_kw['pickle_protocol'] = self._pickle_protocol
        if pyarrow is not None:
            session_kw['arrow_version'] = pyarrow.__version__

        if self._session_id is None:
            resp = self._req_session.post(self._endpoint + '/api/session', data=session_kw)

            if resp.status_code >= 400:
                raise SystemError('Failed to create mars session: ' + resp.reason)
        else:
            resp = self._req_session.get(
                self._endpoint + '/api/session/' + self._session_id, params=session_kw)
            if resp.status_code == 404:
                raise ValueError(f'The session with id = {self._session_id} doesn\'t exist')
            if resp.status_code >= 400:
                raise SystemError('Failed to check mars session.')

        content = json.loads(resp.text)
        self._session_id = content['session_id']
        self._pickle_protocol = content.get('pickle_protocol', pickle.HIGHEST_PROTOCOL)

        # as pyarrow will use pickle.HIGHEST_PROTOCOL to pickle, we need to use
        # SerialType.PICKLE when pickle protocol between client and server
        # does not agree with each other
        if not content.get('arrow_compatible') or self._pickle_protocol != pickle.HIGHEST_PROTOCOL:
            self._serial_type = dataserializer.SerialType.PICKLE

    def _get_tileable_graph_key(self, tileable_key):
        return self._executed_tileables[tileable_key][0]

    def _set_tileable_graph_key(self, tileable, graph_key):
        tileable_key = tileable.key
        tileable_id = tileable.id
        if tileable_key in self._executed_tileables:
            self._executed_tileables[tileable_key][1].add(tileable_id)
        else:
            self._executed_tileables[tileable_key] = graph_key, {tileable_id}

    @staticmethod
    def _handle_json_response(resp, allow_empty=True, raises=True):
        try:
            resp_txt = resp.text
            if allow_empty:
                resp_txt = resp_txt or '{}'
            resp_json = json.loads(resp_txt)
        except json.JSONDecodeError:
            text_part = resp.text if len(resp.text) < 100 else resp.text[:100] + '...'
            raise ResponseMalformed(f'Failed to parse server response. Status={resp.status_code} '
                                    f'Response="{text_part}"')

        if raises and resp.status_code >= 400:
            exc_info = pickle.loads(base64.b64decode(resp_json['exc_info']))
            raise exc_info[1].with_traceback(exc_info[2])
        return resp_json

    def _check_response_finished(self, graph_url, timeout=None):
        import requests
        try:
            resp = self._req_session.get(graph_url, params={'wait_timeout': timeout})
        except requests.ConnectionError as ex:
            err_msg = str(ex)
            if 'ConnectionResetError' in err_msg or 'Connection refused' in err_msg or \
                    'Connection aborted' in err_msg:
                return False
            raise

        if resp.status_code == 504:
            logging.debug('Gateway Time-out, try again')
            return False
        if resp.status_code >= 400:
            raise SystemError(f'Failed to obtain execution status. Code: {resp.status_code}, '
                              f'Reason: {resp.reason}, Content:\n{resp.text}')

        resp_json = self._handle_json_response(resp, raises=False)
        if resp_json['state'] == 'succeeded':
            return True
        elif resp_json['state'] in ('running', 'preparing'):
            return False
        elif resp_json['state'] in ('cancelled', 'cancelling'):
            raise ExecutionInterrupted
        elif resp_json['state'] == 'failed':
            if 'exc_info' in resp_json:
                exc_info = pickle.loads(base64.b64decode(resp_json['exc_info']))
                exc = exc_info[1].with_traceback(exc_info[2])
                raise ExecutionFailed('Graph execution failed.') from exc
            else:
                raise ExecutionFailed('Graph execution failed with unknown reason.')
        raise ExecutionStateUnknown('Unknown graph execution state ' + resp_json['state'])

    def run(self, *tileables, **kw):
        timeout = kw.pop('timeout', -1)
        compose = kw.pop('compose', True)
        fetch = kw.pop('fetch', True)
        name = kw.pop('name', None)
        if kw:
            raise TypeError(f'run got unexpected key arguments {kw!r}')

        # those executed tileables should fetch data directly, submit the others
        run_tileables = [t for t in tileables if t.key not in self._executed_tileables]

        if name is not None:
            if not isinstance(name, (list, tuple)):
                name = [name]
            if len(name) != len(tileables):
                raise TypeError('Name must match execute tileables')
            name = ','.join(name)

        graph = build_tileable_graph(run_tileables, set(self._executed_tileables.keys()))
        targets = [t.key for t in run_tileables]

        if len(graph) > 0:
            targets_join = ','.join(targets)
            session_url = self._endpoint + '/api/session/' + self._session_id
            graph_json = graph.to_json(data_serial_type=self._serial_type, pickle_protocol=self._pickle_protocol)

            resp_json = self._submit_graph(graph_json, targets_join, names=name or '', compose=compose)
            graph_key = resp_json['graph_key']
            graph_url = f'{session_url}/graph/{graph_key}'

            exec_start_time = time.time()
            time_elapsed = 0
            check_interval = options.check_interval
            while timeout <= 0 or time_elapsed < timeout:
                timeout_val = min(check_interval, timeout - time_elapsed) if timeout > 0 else check_interval
                try:
                    if self._check_response_finished(graph_url, timeout_val):
                        break
                except KeyboardInterrupt:
                    resp = self._req_session.delete(graph_url)
                    if resp.status_code >= 400:
                        raise ExecutionNotStopped(
                            f'Failed to stop graph execution. Code: {resp.status_code}, '
                            f'Reason: {resp.reason}, Content:\n{resp.text}')
                finally:
                    time_elapsed = time.time() - exec_start_time

            if 0 < timeout < time.time() - exec_start_time:
                raise TimeoutError

            for t in tileables:
                self._set_tileable_graph_key(t, graph_key)

        if not fetch:
            return
        else:
            return self.fetch(*tileables)

    def _is_executed(self, tileable):
        # if tileble.key in executed tileables
        # or it's a fetch already
        return tileable.key in self._executed_tileables or \
               isinstance(tileable.op, Fetch)

    def fetch(self, *tileables, **kw):
        from ..tensor.indexing import TensorIndex
        from ..dataframe.indexing.iloc import DataFrameIlocGetItem, SeriesIlocGetItem

        timeout = kw.pop('timeout', None)
        if kw:
            raise TypeError(f'fetch got unexpected key arguments {kw!r}')

        results = list()
        for tileable in tileables:
            if tileable.key not in self._executed_tileables and \
                    isinstance(tileable.op, (TensorIndex, DataFrameIlocGetItem, SeriesIlocGetItem)):
                to_fetch_tileable = tileable.inputs[0]
                indexes = tileable.op.indexes
                if not all(isinstance(ind, (slice, Integral)) for ind in indexes):
                    raise ValueError('Only support fetch data slices')
            else:
                to_fetch_tileable = tileable
                indexes = []

            if not self._is_executed(to_fetch_tileable):
                raise ValueError('Cannot fetch the unexecuted tileable')

            key = to_fetch_tileable.key
            indexes_str = json.dumps(Indexes(indexes).to_json(), separators=(',', ':'))

            session_url = f'{self._endpoint}/api/session/{self._session_id}'
            compression_str = ','.join(v.value for v in dataserializer.get_supported_compressions())
            params = dict(compressions=compression_str, slices=indexes_str,
                          serial_type=self._serial_type.value, pickle_protocol=self._pickle_protocol)
            data_url = f'{session_url}/graph/{self._get_tileable_graph_key(key)}/data/{key}'
            resp = self._req_session.get(data_url, params=params, timeout=timeout)
            if resp.status_code >= 400:
                raise ValueError(f'Failed to fetch data from server. Code: {resp.status_code}, '
                                 f'Reason: {resp.reason}, Content:\n{resp.text}')
            result_data = dataserializer.loads(resp.content)
            results.append(sort_dataframe_result(tileable, result_data))
        return results

    @classmethod
    def _process_int_or_dict_argument(cls, argument, name, params):
        if argument is None:
            return
        if not isinstance(argument, dict):
            params[name] = argument
        else:
            params[name] = ','.join(f'{k}={v}' for k, v in argument.items())

    def fetch_tileable_op_logs(self, tileable_op_key, offsets=None, sizes=None):
        url = f'{self._endpoint}/api/session/{self._session_id}/op/{tileable_op_key}/log'
        params = dict()
        self._process_int_or_dict_argument(offsets, 'offsets', params)
        self._process_int_or_dict_argument(sizes, 'sizes', params)
        resp = self._req_session.get(url, params=params)
        if resp.status_code >= 400:
            raise ValueError(f'Failed to fetch log from server. Code: {resp.status_code}, '
                             f'Reason: {resp.reason}, Content:\n{resp.text}')
        return json.loads(resp.content)

    def fetch_log(self, tileables, offsets=None, sizes=None):
        from ..custom_log import fetch

        return fetch(tileables, self, offsets=offsets, sizes=sizes)

    def get_named_tileable_infos(self, name):
        from ..context import TileableInfos

        url = f'{self._endpoint}/api/session/{self._session_id}'
        params = dict(name=name)
        resp = self._req_session.get(url, params=params)
        if resp.status_code >= 400:  # pragma: no cover
            raise ValueError(f'Failed to get tileable key from server. Code: {resp.status_code}, '
                             f'Reason: {resp.reason}, Content:\n{resp.text}')
        tileable_key = self._handle_json_response(resp)['tileable_key']
        nsplits, extra_meta = self._get_tileable_meta(tileable_key)
        shape = tuple(sum(s) for s in nsplits)
        return TileableInfos(tileable_key, shape, extra_meta)

    def create_mutable_tensor(self, name, shape, dtype, fill_value=None, chunk_size=None, *_, **__):
        from ..tensor.utils import create_mutable_tensor
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        tensor_url = f'{session_url}/mutable-tensor/{name}?action=create'
        if not isinstance(dtype, np.dtype):
            dtype = np.dtype(dtype)
        # avoid built-in scalar dtypes are made into one-field record type.
        if dtype.fields:
            dtype_descr = dtype.descr
        else:
            dtype_descr = str(dtype)
        tensor_json = {
            'shape': shape,
            'dtype': dtype_descr,
            'fill_value': fill_value,
            'chunk_size': chunk_size,
        }
        resp = self._req_session.post(tensor_url, json=tensor_json)
        shape, dtype, chunk_size, chunk_keys, chunk_eps = self._handle_json_response(resp)
        return create_mutable_tensor(name, chunk_size, shape, numpy_dtype_from_descr_json(dtype),
                                     chunk_keys, chunk_eps)

    def get_mutable_tensor(self, name):
        from ..tensor.utils import create_mutable_tensor
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        tensor_url = f'{session_url}/mutable-tensor/{name}'
        resp = self._req_session.get(tensor_url)
        shape, dtype, chunk_size, chunk_keys, chunk_eps = self._handle_json_response(resp)
        return create_mutable_tensor(name, chunk_size, shape, numpy_dtype_from_descr_json(dtype),
                                     chunk_keys, chunk_eps)

    def write_mutable_tensor(self, tensor, index, value):
        """
        How to serialize index and value:

        1. process_index and serialize it as json
        2. the payload of POST request:

            * a int64 value indicate the size of index json
            * ascii-encoded bytes of index json
            * pyarrow serialized bytes of `value`
        """
        from ..tensor.core import Indexes
        from ..serialize import dataserializer

        index = Indexes(_indexes=index)
        index_bytes = json.dumps(index.to_json()).encode('ascii')
        bio = BytesIO()
        bio.write(np.int64(len(index_bytes)).tobytes())
        bio.write(index_bytes)
        dataserializer.dump(value, bio)

        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        tensor_url = f'{session_url}/mutable-tensor/{tensor.name}'
        resp = self._req_session.put(tensor_url, data=bio.getvalue(),
                                     headers={'Content-Type': 'application/octet-stream'})
        self._handle_json_response(resp)

    def seal(self, tensor):
        from ..tensor.utils import create_fetch_tensor
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        tensor_url = f'{session_url}/mutable-tensor/{tensor.name}?action=seal'
        resp = self._req_session.post(tensor_url)
        graph_key_hex, tileable_key, tensor_id, tensor_meta = self._handle_json_response(resp)
        self._executed_tileables[tileable_key] = uuid.UUID(graph_key_hex), {tensor_id}

        # # Construct Tensor on the fly.
        shape, dtype, chunk_size, chunk_keys, _ = tensor_meta
        return create_fetch_tensor(chunk_size, shape, numpy_dtype_from_descr_json(dtype),
                                   tensor_key=tileable_key, chunk_keys=chunk_keys)

    def _get_tileable_nsplits(self, tileable_key):
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        graph_key = self._get_tileable_graph_key(tileable_key)
        url = f'{session_url}/graph/{graph_key}/data/{tileable_key}?type=nsplits'
        resp = self._req_session.get(url)
        new_nsplits = self._handle_json_response(resp)
        return new_nsplits

    def _get_tileable_meta(self, tileable_key):
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        graph_key = self._get_tileable_graph_key(tileable_key)
        url = f'{session_url}/graph/{graph_key}/data/{tileable_key}?type=meta'
        resp = self._req_session.get(url)
        meta = self._handle_json_response(resp)
        return pickle.loads(base64.b64decode(meta))  # nosec

    def _update_tileable_shape(self, tileable):
        tileable_key = tileable.key
        new_nsplits = self._get_tileable_nsplits(tileable_key)
        tileable._update_shape(tuple(sum(nsplit) for nsplit in new_nsplits))
        tileable.nsplits = new_nsplits

    def decref(self, *keys):
        for tileable_key, tileable_id in keys:
            if tileable_key not in self._executed_tileables:
                continue
            graph_key, ids = self._executed_tileables[tileable_key]

            if tileable_id in ids:
                ids.remove(tileable_id)
                # for those same key tileables, do decref only when all those tileables are garbage collected
                if len(ids) != 0:
                    continue
                self.delete_data(tileable_key)

    def delete_data(self, tileable_key, wait=False):
        if tileable_key not in self._executed_tileables:
            return
        graph_key, _ids = self._executed_tileables[tileable_key]
        data_url = f'{self._endpoint}/api/session/{self._session_id}/graph/{graph_key}' \
                   f'/data/{tileable_key}?wait={1 if wait else 0}'
        self._req_session.delete(data_url)
        self._executed_tileables.pop(tileable_key, None)

    def stop(self, graph_key):
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        graph_url = session_url + '/graph/' + graph_key
        resp = self._req_session.delete(graph_url)
        if resp.status_code >= 400:
            raise SystemError(f'Failed to stop graph execution. Code: {resp.status_code}, '
                              f'Reason: {resp.reason}, Content:\n{resp.text}')

    def _submit_graph(self, graph_json, targets, names=None, compose=True):
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        resp = self._req_session.post(session_url + '/graph', dict(
            graph=json.dumps(graph_json),
            target=targets,
            names=names,
            compose='1' if compose else '0'
        ))
        return self._handle_json_response(resp)

    def get_graph_states(self):
        resp = self._req_session.get(f'{self._endpoint}/api/session/{self._session_id}/graph')
        return self._handle_json_response(resp)

    def close(self):
        session_url = f'{self._endpoint}/api/session/{self._session_id}'
        for key in list(self._executed_tileables.keys()):
            self.delete_data(key, wait=True)
        resp = self._req_session.delete(session_url)
        if resp.status_code >= 400:
            raise SystemError('Failed to close mars session.')

    def check_service_ready(self, timeout=1):
        import requests
        try:
            resp = self._req_session.get(self._endpoint + '/api', timeout=timeout)
        except (requests.ConnectionError, requests.Timeout):
            return False
        if resp.status_code >= 400:
            return False
        return True

    def count_workers(self):
        resp = self._req_session.get(self._endpoint + '/api/worker?action=count', timeout=1)
        return self._handle_json_response(resp)

    def rescale_workers(self, new_scale, min_workers=None, wait=True, timeout=None):
        data = json.dumps(dict(new_scale=new_scale, min_workers=min_workers))
        wait_req = 1 if wait else 0
        resp = self._req_session.patch(f'{self._endpoint}/api/worker?action=count&wait={wait_req}',
                                       data, timeout=timeout)
        return self._handle_json_response(resp)

    def get_workers_meta(self):
        resp = self._req_session.get(self._endpoint + '/api/worker', timeout=1)
        return self._handle_json_response(resp)

    def get_task_count(self):
        resp = self._req_session.get(f'{self._endpoint}/api/session/{self._session_id}/graph')
        return len(self._handle_json_response(resp))

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utils import check_consistent_length
from ..utils import make_import_error_func
from .core import LGBMScikitLearnBase, LGBMModelType
from ._train import train
from ._predict import predict_base

try:
    import lightgbm
except ImportError:
    lightgbm = None


LGBMRanker = make_import_error_func('lightgbm')
if lightgbm:
    class LGBMRanker(LGBMScikitLearnBase, lightgbm.LGBMRanker):
        def fit(self, X, y, sample_weight=None, init_score=None, group=None, eval_set=None,
                eval_sample_weight=None, eval_init_score=None,
                session=None, run_kwargs=None, **kwargs):
            check_consistent_length(X, y, session=session, run_kwargs=run_kwargs)
            params = self.get_params(True)
            model = train(params, self._wrap_train_tuple(X, y, sample_weight, init_score),
                          eval_sets=self._wrap_eval_tuples(eval_set, eval_sample_weight, eval_init_score),
                          group=group, model_type=LGBMModelType.RANKER,
                          session=session, run_kwargs=run_kwargs, **kwargs)

            self.set_params(**model.get_params())
            self._copy_extra_params(model, self)
            return self

        def predict(self, X, **kw):
            session = kw.pop('session', None)
            run_kwargs = kw.pop('run_kwargs', None)
            return predict_base(self, X, session=session, run_kwargs=run_kwargs, **kw)

        def to_local(self):
            model = lightgbm.LGBMRanker(**self.get_params())
            self._copy_extra_params(self, model)
            return model



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/operand.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/operand.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#mars/serialize/protos/operand.proto\x1a!mars/serialize/protos/value.proto\"\x80\x01\n\nOperandDef\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12#\n\x04\x61ttr\x18\x03 \x03(\x0b\x32\x15.OperandDef.AttrEntry\x1a\x33\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])




_OPERANDDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='OperandDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='OperandDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='OperandDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=203,
)

_OPERANDDEF = _descriptor.Descriptor(
  name='OperandDef',
  full_name='OperandDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OperandDef.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='OperandDef.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attr', full_name='OperandDef.attr', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_OPERANDDEF_ATTRENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=203,
)

_OPERANDDEF_ATTRENTRY.fields_by_name['value'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OPERANDDEF_ATTRENTRY.containing_type = _OPERANDDEF
_OPERANDDEF.fields_by_name['attr'].message_type = _OPERANDDEF_ATTRENTRY
DESCRIPTOR.message_types_by_name['OperandDef'] = _OPERANDDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OperandDef = _reflection.GeneratedProtocolMessageType('OperandDef', (_message.Message,), {

  'AttrEntry' : _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), {
    'DESCRIPTOR' : _OPERANDDEF_ATTRENTRY,
    '__module__' : 'mars.serialize.protos.operand_pb2'
    # @@protoc_insertion_point(class_scope:OperandDef.AttrEntry)
    })
  ,
  'DESCRIPTOR' : _OPERANDDEF,
  '__module__' : 'mars.serialize.protos.operand_pb2'
  # @@protoc_insertion_point(class_scope:OperandDef)
  })
_sym_db.RegisterMessage(OperandDef)
_sym_db.RegisterMessage(OperandDef.AttrEntry)


_OPERANDDEF_ATTRENTRY._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .... import opcodes as OperandDef
from .... import tensor as mt
from ....serialize import KeyField, BoolField
from ....tensor.core import TensorOrder
from ....tiles import TilesError
from ....utils import recursive_tile, check_chunks_unknown_shape
from ...utils import check_array
from ...utils.extmath import row_norms
from .core import PairwiseDistances


class EuclideanDistances(PairwiseDistances):
    _op_type_ = OperandDef.PAIRWISE_EUCLIDEAN_DISTANCES

    _x = KeyField('X')
    _y = KeyField('Y')
    _x_norm_squared = KeyField('X_norm_squared')
    _y_norm_squared = KeyField('Y_norm_squared')
    _squared = BoolField('squared')

    def __init__(self, x=None, y=None, x_norm_squared=None, y_norm_squared=None,
                 squared=None, dtype=None, gpu=None, **kw):
        super().__init__(_x=x, _y=y, _x_norm_squared=x_norm_squared,
                         _y_norm_squared=y_norm_squared, _squared=squared,
                         _dtype=dtype, _gpu=gpu, **kw)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def x_norm_squared(self):
        return self._x_norm_squared

    @property
    def y_norm_squared(self):
        return self._y_norm_squared

    @property
    def squared(self):
        return self._squared

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        input_iter = iter(self._inputs)
        self._x = next(input_iter)
        if self._y is not None:
            self._y = next(input_iter)
        if self._x_norm_squared is not None:
            self._x_norm_squared = next(input_iter)
        if self._y_norm_squared is not None:
            self._y_norm_squared = next(input_iter)

    def __call__(self, X, Y=None, Y_norm_squared=None, X_norm_squared=None):
        # If norms are passed as float32, they are unused. If arrays are passed as
        # float32, norms needs to be recomputed on upcast chunks.
        # TODO: use a float64 accumulator in row_norms to avoid the latter.
        if X_norm_squared is not None:
            XX = check_array(X_norm_squared)
            if XX.shape == (1, X.shape[0]):
                XX = XX.T
            elif XX.shape != (X.shape[0], 1):
                raise ValueError(
                    "Incompatible dimensions for X and X_norm_squared")
            if XX.dtype == np.float32:
                XX = self._x_norm_squared = None
        else:
            XX = None

        if X is Y and XX is not None:
            # shortcut in the common case euclidean_distances(X, X)
            YY = XX.T
        elif Y_norm_squared is not None:
            YY = mt.atleast_2d(Y_norm_squared)

            if YY.shape != (1, Y.shape[0]):
                raise ValueError(
                    "Incompatible dimensions for Y and Y_norm_squared")
            if YY.dtype == np.float32:
                YY = self._y_norm_squared = None
        else:
            YY = None

        inputs = [X, Y]
        if XX is not None:
            inputs.append(XX)
        if YY is not None:
            inputs.append(YY)
        return self.new_tensor(inputs, shape=(X.shape[0], Y.shape[0]),
                               order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        X, Y = op.x, op.y
        out = op.outputs[0]

        if X.dtype == np.float32:
            check_chunks_unknown_shape([X, Y], TilesError)
            # rechunk
            new_nsplit = max(max(X.nsplits[0]) // 2, 1)
            X = recursive_tile(X.rechunk({0: new_nsplit}).astype(np.float64))
            if Y is not X:
                new_nsplit = max(max(Y.nsplits[0]) // 2, 1)
                Y = recursive_tile(Y.rechunk({0: new_nsplit}).astype(np.float64))

        XX = op.x_norm_squared
        if XX is None:
            XX = row_norms(X, squared=True)[:, np.newaxis]
        YY = op.y_norm_squared
        if YY is None:
            YY = row_norms(Y, squared=True)[np.newaxis, :]

        X, Y = cls._adjust_chunk_sizes(X, Y, out)

        distances = -2 * X.dot(Y.T)
        if distances.issparse():
            distances = distances.todense()
        distances += XX
        distances += YY
        distances = mt.maximum(distances, 0)

        if X is Y or X.key == Y.key:
            mt.fill_diagonal(distances, 0)

        distances = distances if op.squared else mt.sqrt(distances)
        distances = distances.astype(out.dtype, copy=False)
        return [recursive_tile(distances)]


def euclidean_distances(X, Y=None, Y_norm_squared=None, squared=False,
                        X_norm_squared=None):
    """
    Considering the rows of X (and Y=X) as vectors, compute the
    distance matrix between each pair of vectors.

    For efficiency reasons, the euclidean distance between a pair of row
    vector x and y is computed as::

        dist(x, y) = sqrt(dot(x, x) - 2 * dot(x, y) + dot(y, y))

    This formulation has two advantages over other ways of computing distances.
    First, it is computationally efficient when dealing with sparse data.
    Second, if one argument varies but the other remains unchanged, then
    `dot(x, x)` and/or `dot(y, y)` can be pre-computed.

    However, this is not the most precise way of doing this computation, and
    the distance matrix returned by this function may not be exactly
    symmetric as required by, e.g., ``scipy.spatial.distance`` functions.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape (n_samples_1, n_features)

    Y : {array-like, sparse matrix}, shape (n_samples_2, n_features)

    Y_norm_squared : array-like, shape (n_samples_2, ), optional
        Pre-computed dot-products of vectors in Y (e.g.,
        ``(Y**2).sum(axis=1)``)
        May be ignored in some cases, see the note below.

    squared : boolean, optional
        Return squared Euclidean distances.

    X_norm_squared : array-like, shape = [n_samples_1], optional
        Pre-computed dot-products of vectors in X (e.g.,
        ``(X**2).sum(axis=1)``)
        May be ignored in some cases, see the note below.

    Notes
    -----
    To achieve better accuracy, `X_norm_squared` and `Y_norm_squared` may be
    unused if they are passed as ``float32``.

    Returns
    -------
    distances : tensor, shape (n_samples_1, n_samples_2)

    Examples
    --------
    >>> from mars.learn.metrics.pairwise import euclidean_distances
    >>> X = [[0, 1], [1, 1]]
    >>> # distance between rows of X
    >>> euclidean_distances(X, X).execute()
    array([[0., 1.],
           [1., 0.]])
    >>> # get distance to origin
    >>> euclidean_distances(X, [[0, 0]]).execute()
    array([[1.        ],
           [1.41421356]])

    See also
    --------
    paired_distances : distances betweens pairs of elements of X and Y.
    """
    if X.dtype == np.float32:
        if Y is None:
            dtype = X.dtype
        elif Y.dtype == np.float32:
            dtype = np.float32
        else:
            dtype = np.float64
    else:
        dtype = np.float64

    X, Y = EuclideanDistances.check_pairwise_arrays(X, Y)
    op = EuclideanDistances(x=X, y=Y, x_norm_squared=X_norm_squared,
                            y_norm_squared=Y_norm_squared, squared=squared,
                            dtype=np.dtype(dtype))
    return op(X, Y=Y, Y_norm_squared=Y_norm_squared,
              X_norm_squared=X_norm_squared)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import uuid

import numpy as np

from ..worker.transfer import ReceiverManagerActor
from ..utils import log_unhandled, tokenize
from .utils import SchedulerActor


class MutableTensorActor(SchedulerActor):
    """
    Actor handling processing of a Mars mutable tensor.
    """
    @staticmethod
    def gen_uid(session_id, name):
        return f's:0:mutable-tensor${session_id}${name}'

    def __init__(self, session_id, name, shape, dtype, graph_key, fill_value=None, chunk_size=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._session_id = session_id
        self._name = name
        self._shape = shape
        if isinstance(dtype, np.dtype):
            self._dtype = dtype
        else:
            self._dtype = np.dtype(dtype)
        self._graph_key = graph_key
        self._chunk_size = chunk_size
        self._fill_value = fill_value
        self._tensor = None
        self._sealed = False
        self._chunk_map = defaultdict(lambda: [])
        self._chunk_to_endpoint = dict()
        self._record_type = np.dtype([("index", np.uint32),
                                      ("ts", np.dtype('datetime64[ns]')),
                                      ("value", self._dtype)])
        self._resource_ref = None

    @log_unhandled
    def post_create(self):
        from .resource import ResourceActor
        from ..tensor.core import MutableTensor, MutableTensorData
        from ..tensor.utils import create_fetch_tensor

        super().post_create()
        self.set_cluster_info_ref()
        self._resource_ref = self.ctx.actor_ref(ResourceActor.default_uid())

        tensor = create_fetch_tensor(self._chunk_size, self._shape, self._dtype)
        endpoints = self._resource_ref.select_worker(size=len(tensor.chunks))

        for chunk, endpoint in zip(tensor.chunks, endpoints):
            self._chunk_to_endpoint[chunk.key] = endpoint

        self._tensor = MutableTensor(data=MutableTensorData(
            name=self._name, op=None, shape=self._shape, dtype=tensor.dtype,
            nsplits=tensor.nsplits, key=tensor.key, chunks=tensor.chunks,
            chunk_eps=endpoints))

    def pre_destroy(self):
        self.unset_cluster_info_ref()

    def tensor_meta(self):
        # avoid built-in scalar dtypes are made into one-field record type.
        if self._dtype.fields:
            dtype_descr = self._dtype.descr
        else:
            dtype_descr = str(self._dtype)
        chunk_keys = [c.key for c in self._tensor.chunks]
        chunk_eps = [self._chunk_to_endpoint[key] for key in chunk_keys]
        return self._shape, dtype_descr, self._chunk_size, chunk_keys, chunk_eps

    def tensor_key(self):
        return self._tensor.key

    def sealed(self):
        return self._sealed

    @log_unhandled
    def read(self, tensor_index):
        raise NotImplementedError

    @log_unhandled
    def write(self, index, value):
        chunk_records_to_send = self._tensor._do_write(index, value)
        self._send_chunk_records(chunk_records_to_send)

    @log_unhandled
    def append_chunk_records(self, chunk_records):
        for chunk_key, record_chunk_key in chunk_records:
            self._chunk_map[chunk_key].append(record_chunk_key)

    @log_unhandled
    def seal(self):
        from ..worker.seal import SealActor
        self._sealed = True

        # dump current buffers to worker
        chunk_records_to_send = self._tensor._do_flush()
        self._send_chunk_records(chunk_records_to_send)

        endpoint_to_sealer = dict()

        # consolidate chunks
        for chunk in self._tensor.chunks:
            ep = self._chunk_to_endpoint[chunk.key]
            if ep in endpoint_to_sealer:
                sealer_ref = endpoint_to_sealer[ep]
            else:
                sealer_uid = SealActor.gen_uid(self._session_id, chunk.key)
                sealer_ref = endpoint_to_sealer[ep] = self.ctx.create_actor(SealActor, uid=sealer_uid, address=ep)
            sealer_ref.seal_chunk(self._session_id, self._graph_key,
                                  chunk.key, self._chunk_map[chunk.key],
                                  chunk.shape, self._record_type, self._dtype, self._fill_value)

        for sealer_ref in endpoint_to_sealer.values():
            sealer_ref.destroy()
        # return the hex of self._graph_key since UUID is not json serializable.
        return self._graph_key.hex, self._tensor.key, self._tensor.id, self.tensor_meta()

    @log_unhandled
    def _send_chunk_records(self, chunk_records_to_send):
        from ..worker.transfer import put_remote_chunk

        chunk_records = []

        for chunk_key, ep, records in chunk_records_to_send:
            record_chunk_key = tokenize(chunk_key, uuid.uuid4().hex)
            # send record chunk
            receiver_manager_ref = self.ctx.actor_ref(ReceiverManagerActor.default_uid(), address=ep)
            put_remote_chunk(self._session_id, record_chunk_key, records, receiver_manager_ref)
            chunk_records.append((chunk_key, record_chunk_key))

        # register the record chunks
        self.append_chunk_records(chunk_records)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from collections import defaultdict

from .. import promise
from ..config import options
from ..executor import Executor
from ..serialize import dataserializer
from ..utils import to_str, deserialize_graph, log_unhandled, calc_data_size, \
    get_chunk_shuffle_key, build_exc_info
from ..context import DistributedDictContext
from .events import EventContext, EventCategory, EventLevel, ProcedureEventType
from .storage import DataStorageDevice
from .utils import WorkerActor, concat_operand_keys, get_chunk_key, build_quota_key

logger = logging.getLogger(__name__)


class BaseCalcActor(WorkerActor):
    _slot_name = None
    _calc_event_type = None
    _calc_source_devices = None
    _calc_intermediate_device = None
    _calc_dest_devices = None

    def __init__(self):
        super().__init__()
        self._remove_intermediate = self._calc_intermediate_device not in self._calc_dest_devices

        self._dispatch_ref = None
        self._mem_quota_ref = None
        self._events_ref = None
        self._status_ref = None
        self._execution_ref = None
        self._resource_ref = None

        self._executing_set = set()
        self._marked_as_destroy = False

        self._execution_pool = None
        self._n_cpu = None

    def post_create(self):
        super().post_create()

        from .dispatcher import DispatchActor
        from .execution import ExecutionActor
        from .events import EventsActor
        from .quota import MemQuotaActor
        from .status import StatusActor
        from ..scheduler.resource import ResourceActor

        self._mem_quota_ref = self.promise_ref(MemQuotaActor.default_uid())
        self._dispatch_ref = self.promise_ref(DispatchActor.default_uid())
        self._dispatch_ref.register_free_slot(self.uid, self._slot_name)

        status_ref = self.ctx.actor_ref(StatusActor.default_uid())
        self._status_ref = status_ref if self.ctx.has_actor(status_ref) else None

        self._execution_ref = self.ctx.actor_ref(ExecutionActor.default_uid())
        if not self.ctx.has_actor(self._execution_ref):
            self._execution_ref = None

        self._events_ref = self.ctx.actor_ref(EventsActor.default_uid())
        if not self.ctx.has_actor(self._events_ref):
            self._events_ref = None

        self._resource_ref = self.get_actor_ref(ResourceActor.default_uid())

        self._execution_pool = self.ctx.threadpool(1)

    def pre_destroy(self):
        logger.debug('Destroying actor %s', self.uid)
        self._dispatch_ref.remove_slot(self.uid, self._slot_name, _tell=True, _wait=False)
        super().pre_destroy()

    @staticmethod
    def _get_keys_to_fetch(graph):
        from ..operands import Fetch, FetchShuffle
        fetch_keys = set()
        exclude_fetch_keys = set()
        for chunk in graph:
            if isinstance(chunk.op, Fetch):
                fetch_keys.add(chunk.op.to_fetch_key or chunk.key)
            elif isinstance(chunk.op, FetchShuffle):
                shuffle_key = get_chunk_shuffle_key(graph.successors(chunk)[0])
                for k in chunk.op.to_fetch_keys:
                    fetch_keys.add((k, shuffle_key))
            else:
                for inp, is_pure_dep in zip(chunk.inputs, chunk.op.pure_depends):
                    if is_pure_dep:
                        exclude_fetch_keys.add(inp.key)
        return list(fetch_keys - exclude_fetch_keys)

    def _make_quotas_local(self, session_id, graph_key, data_keys, process_quota=False):
        old_keys, new_keys = [], []
        for k in data_keys:
            old_keys.append(build_quota_key(session_id, k, owner=graph_key))
            new_keys.append(build_quota_key(session_id, k, owner=self.proc_id))
        self._mem_quota_ref.alter_allocations(
            old_keys, new_keys=new_keys, process_quota=process_quota)
        return new_keys

    def _release_local_quotas(self, session_id, data_keys):
        self._mem_quota_ref.release_quotas(
            [build_quota_key(session_id, k, owner=self.proc_id) for k in data_keys],
            _tell=True, _wait=False)

    def _fetch_keys_to_process(self, session_id, keys_to_fetch):
        context_dict = dict()
        storage_client = self.storage_client

        if not keys_to_fetch:
            return promise.finished(context_dict)

        def _handle_loaded(objs):
            try:
                data_locations = storage_client.get_data_locations(session_id, keys_to_fetch)
                shared_quota_keys = []
                inproc_keys = []
                inproc_quota_keys = []

                context_dict.update(zip(keys_to_fetch, objs))
                for k, locations in zip(keys_to_fetch, data_locations):
                    quota_key = build_quota_key(session_id, k, owner=self.proc_id)
                    if (self.proc_id, DataStorageDevice.PROC_MEMORY) not in locations:
                        shared_quota_keys.append(quota_key)
                    else:
                        inproc_keys.append(k)
                        inproc_quota_keys.append(quota_key)

                if shared_quota_keys:
                    self._mem_quota_ref.hold_quotas(shared_quota_keys, _tell=True)
                if inproc_keys:
                    self._mem_quota_ref.hold_quotas(inproc_quota_keys, _tell=True)
                    if self._remove_intermediate:
                        storage_client.delete(session_id, inproc_keys, [self._calc_intermediate_device])
            finally:
                objs[:] = []

        def _handle_load_fail(*exc):
            if self._remove_intermediate:
                storage_client.delete(session_id, keys_to_fetch, [self._calc_intermediate_device])
            context_dict.clear()
            raise exc[1].with_traceback(exc[2])

        return storage_client.get_objects(session_id, keys_to_fetch, self._calc_source_devices) \
            .then(_handle_loaded, _handle_load_fail).then(lambda *_: context_dict)

    def _get_n_cpu(self):
        if self._n_cpu is None:
            self._n_cpu = len(self._dispatch_ref.get_slots('cpu'))
        return self._n_cpu

    @log_unhandled
    def _calc_results(self, session_id, graph_key, graph, context_dict, chunk_targets):
        _, op_name = concat_operand_keys(graph, '_')

        logger.debug('Start calculating operand %s in %s.', graph_key, self.uid)
        start_time = time.time()

        for chunk in graph:
            for inp, is_pure_dep in zip(chunk.inputs, chunk.op.pure_depends):
                if is_pure_dep:
                    context_dict[inp.key] = None

        local_context_dict = DistributedDictContext(
            self.get_scheduler(self.default_uid()), session_id, actor_ctx=self.ctx,
            address=self.address, n_cpu=self._get_n_cpu(), is_distributed=True,
            resource_ref=self._resource_ref)
        local_context_dict['_actor_cls'] = type(self)
        local_context_dict['_actor_uid'] = self.uid
        local_context_dict['_op_key'] = graph_key
        local_context_dict.update(context_dict)
        context_dict.clear()

        if self._execution_ref:
            self._execution_ref.deallocate_scheduler_resource(
                session_id, graph_key, delay=0.5, _tell=True, _wait=False)

        # start actual execution
        executor = Executor(storage=local_context_dict)
        with EventContext(self._events_ref, EventCategory.PROCEDURE, EventLevel.NORMAL,
                          self._calc_event_type, self.uid):
            self._execution_pool.submit(executor.execute_graph, graph,
                                        chunk_targets, retval=False).result()

        end_time = time.time()

        # collect results
        result_keys = []
        result_values, result_sizes, result_shapes = [], [], []
        collected_chunk_keys = set()
        for k in list(local_context_dict.keys()):
            v = local_context_dict[k]
            if isinstance(k, tuple):
                k = tuple(to_str(i) for i in k)
            else:
                k = to_str(k)

            chunk_key = get_chunk_key(k)
            if chunk_key in chunk_targets:
                result_keys.append(k)
                if self._calc_intermediate_device in self._calc_dest_devices:
                    result_values.append(v)
                    result_sizes.append(calc_data_size(v))
                else:
                    result_values.append(dataserializer.serialize(v))
                    result_sizes.append(result_values[-1].total_bytes)
                result_shapes.append(getattr(v, 'shape', None))
                collected_chunk_keys.add(chunk_key)

            local_context_dict.pop(k)

        # check if all targets generated
        if any(k not in collected_chunk_keys for k in chunk_targets):
            raise KeyError([k for k in chunk_targets if k not in collected_chunk_keys])

        # adjust sizes in allocation
        apply_allocs = defaultdict(lambda: 0)
        for k, size in zip(result_keys, result_sizes):
            apply_allocs[get_chunk_key(k)] += size

        apply_alloc_quota_keys, apply_alloc_sizes = [], []
        for k, v in apply_allocs.items():
            apply_alloc_quota_keys.append(build_quota_key(session_id, k, owner=self.proc_id))
            apply_alloc_sizes.append(v)
        self._mem_quota_ref.alter_allocations(apply_alloc_quota_keys, apply_alloc_sizes, _tell=True, _wait=False)
        self._mem_quota_ref.hold_quotas(apply_alloc_quota_keys, _tell=True)

        if self._status_ref:
            self._status_ref.update_mean_stats(
                'calc_speed.' + op_name, sum(apply_alloc_sizes) * 1.0 / (end_time - start_time),
                _tell=True, _wait=False)

        return self.storage_client.put_objects(
            session_id, result_keys, result_values, [self._calc_intermediate_device],
            sizes=result_sizes, shapes=result_shapes) \
            .then(lambda *_: result_keys)

    @promise.reject_on_exception
    @log_unhandled
    def calc(self, session_id, graph_key, ser_graph, chunk_targets, callback):
        """
        Do actual calculation. This method should be called when all data
        is available (i.e., either in shared cache or in memory)
        :param session_id: session id
        :param graph_key: key of executable graph
        :param ser_graph: serialized executable graph
        :param chunk_targets: keys of target chunks
        :param callback: promise callback, returns the uid of InProcessCacheActor
        """
        self._executing_set.add(graph_key)
        graph = deserialize_graph(ser_graph)
        chunk_targets = set(chunk_targets)
        keys_to_fetch = self._get_keys_to_fetch(graph)

        self._make_quotas_local(session_id, graph_key, keys_to_fetch + list(chunk_targets),
                                process_quota=True)

        def _start_calc(context_dict):
            return self._calc_results(session_id, graph_key, graph, context_dict, chunk_targets)

        def _finalize(keys, exc_info):
            if not self._marked_as_destroy:
                self._dispatch_ref.register_free_slot(self.uid, self._slot_name, _tell=True, _wait=False)

            if not exc_info:
                self.tell_promise(callback, keys)
            else:
                try:
                    self.tell_promise(callback, *exc_info, _accept=False)
                except:
                    self.tell_promise(callback, *build_exc_info(
                        SystemError, f'Failed to send errors to scheduler, type: {exc_info[0].__name__}, '
                                     f'message: {str(exc_info[1])}'), _accept=False)
                    raise

            keys_to_release = [k for k in keys_to_fetch if get_chunk_key(k) not in chunk_targets]
            if exc_info:
                keys_to_release.extend(chunk_targets)

            if self._remove_intermediate:
                keys_to_delete = keys_to_release
            else:
                keys_to_delete = []

            if keys_to_delete:
                self.storage_client.delete(session_id, keys_to_delete, [self._calc_intermediate_device])
            logger.debug('Finish calculating operand %s.', graph_key)

        return self._fetch_keys_to_process(session_id, keys_to_fetch) \
            .then(lambda context_dict: _start_calc(context_dict)) \
            .then(lambda keys: _finalize(keys, None), lambda *exc_info: _finalize(None, exc_info))

    @promise.reject_on_exception
    @log_unhandled
    def store_results(self, session_id, graph_key, keys_to_store, data_attrs, callback):
        logger.debug('Start dumping keys for graph %s', graph_key)
        from ..scheduler.chunkmeta import WorkerMeta

        storage_client = self.storage_client
        store_keys, store_metas = [], []

        if data_attrs is None:
            data_attrs = storage_client.get_data_attrs(session_id, keys_to_store)

        for k, attr in zip(keys_to_store, data_attrs):
            if attr is None or isinstance(k, tuple):
                continue
            store_keys.append(k)
            store_metas.append(WorkerMeta(attr.size, attr.shape, (self.address,)))
        meta_future = self.get_meta_client().batch_set_chunk_meta(
            session_id, store_keys, store_metas, _wait=False)

        def _delete_keys(*_):
            logger.debug('Finish dumping keys for graph %s', graph_key)
            if self._remove_intermediate:
                storage_client.delete(
                    session_id, keys_to_store, [self._calc_intermediate_device], _tell=True)
            self._release_local_quotas(session_id, keys_to_store)

        def _finalize(*_):
            self._executing_set.remove(graph_key)
            if self._marked_as_destroy and not self._executing_set:
                # delay destroying to allow requests to enter
                self.ref().destroy_self(_tell=True, _delay=0.5)

        dest_devices = self._calc_dest_devices \
            if not isinstance(keys_to_store[0], tuple) or not options.worker.write_shuffle_to_disk \
            else [DataStorageDevice.DISK]
        return storage_client.copy_to(session_id, keys_to_store, dest_devices) \
            .then(_delete_keys) \
            .then(lambda *_: meta_future.result()) \
            .then(lambda *_: self.tell_promise(callback),
                  lambda *exc: self.tell_promise(callback, *exc, _accept=False)) \
            .then(_finalize)

    def remove_cache(self, session_id, keys):
        self.storage_client.delete(
            session_id, keys, [self._calc_intermediate_device], _tell=True)

    def destroy_self(self):
        if self._executing_set:
            return

        from .daemon import WorkerDaemonActor
        daemon_ref = self.ctx.actor_ref(WorkerDaemonActor.default_uid())
        if self.ctx.has_actor(daemon_ref):
            daemon_ref.destroy_actor(self.ref(), _tell=True)
        else:
            self.ctx.destroy_actor(self.ref())

    @log_unhandled
    def mark_destroy(self):
        self._marked_as_destroy = True
        self._dispatch_ref.remove_slot(self.uid, self._slot_name)
        if not self._executing_set:
            # delay destroying to allow requests to enter
            self.ref().destroy_self(_tell=True, _delay=0.5)


class CpuCalcActor(BaseCalcActor):
    _slot_name = 'cpu'
    _calc_event_type = ProcedureEventType.CPU_CALC
    if options.vineyard.enabled:
        shared_memory_device = DataStorageDevice.VINEYARD  # pragma: no cover
    else:
        shared_memory_device = DataStorageDevice.SHARED_MEMORY

    # PROC_MEMORY must come first to avoid loading to shared storage
    # which will easily cause errors
    _calc_source_devices = (DataStorageDevice.PROC_MEMORY, shared_memory_device)
    _calc_intermediate_device = DataStorageDevice.PROC_MEMORY
    _calc_dest_devices = (shared_memory_device, DataStorageDevice.DISK)


class CudaCalcActor(BaseCalcActor):
    _slot_name = 'cuda'
    _calc_event_type = ProcedureEventType.GPU_CALC
    _calc_source_devices = (DataStorageDevice.CUDA, )
    _calc_intermediate_device = DataStorageDevice.CUDA
    _calc_dest_devices = (DataStorageDevice.CUDA, )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorRealIFFTNMixin, validate_fftn, TensorRealFFTN


class TensorIRFFTN(TensorRealFFTN, TensorRealIFFTNMixin):
    _op_type_ = OperandDef.IRFFTN

    def __init__(self, shape=None, axes=None, norm=None, dtype=None, **kw):
        super().__init__(_shape=shape, _axes=axes, _norm=norm, _dtype=dtype, **kw)


def irfftn(a, s=None, axes=None, norm=None):
    """
    Compute the inverse of the N-dimensional FFT of real input.

    This function computes the inverse of the N-dimensional discrete
    Fourier Transform for real input over any number of axes in an
    M-dimensional tensor by means of the Fast Fourier Transform (FFT).  In
    other words, ``irfftn(rfftn(a), a.shape) == a`` to within numerical
    accuracy. (The ``a.shape`` is necessary like ``len(a)`` is for `irfft`,
    and for the same reason.)

    The input should be ordered in the same way as is returned by `rfftn`,
    i.e. as for `irfft` for the final transformation axis, and as for `ifftn`
    along all the other axes.

    Parameters
    ----------
    a : array_like
        Input tensor.
    s : sequence of ints, optional
        Shape (length of each transformed axis) of the output
        (``s[0]`` refers to axis 0, ``s[1]`` to axis 1, etc.). `s` is also the
        number of input points used along this axis, except for the last axis,
        where ``s[-1]//2+1`` points of the input are used.
        Along any axis, if the shape indicated by `s` is smaller than that of
        the input, the input is cropped.  If it is larger, the input is padded
        with zeros. If `s` is not given, the shape of the input along the
        axes specified by `axes` is used.
    axes : sequence of ints, optional
        Axes over which to compute the inverse FFT. If not given, the last
        `len(s)` axes are used, or all axes if `s` is also not specified.
        Repeated indices in `axes` means that the inverse transform over that
        axis is performed multiple times.
    norm : {None, "ortho"}, optional
        Normalization mode (see `mt.fft`). Default is None.

    Returns
    -------
    out : Tensor
        The truncated or zero-padded input, transformed along the axes
        indicated by `axes`, or by a combination of `s` or `a`,
        as explained in the parameters section above.
        The length of each transformed axis is as given by the corresponding
        element of `s`, or the length of the input in every axis except for the
        last one if `s` is not given.  In the final transformed axis the length
        of the output when `s` is not given is ``2*(m-1)`` where ``m`` is the
        length of the final transformed axis of the input.  To get an odd
        number of output points in the final axis, `s` must be specified.

    Raises
    ------
    ValueError
        If `s` and `axes` have different length.
    IndexError
        If an element of `axes` is larger than than the number of axes of `a`.

    See Also
    --------
    rfftn : The forward n-dimensional FFT of real input,
            of which `ifftn` is the inverse.
    fft : The one-dimensional FFT, with definitions and conventions used.
    irfft : The inverse of the one-dimensional FFT of real input.
    irfft2 : The inverse of the two-dimensional FFT of real input.

    Notes
    -----
    See `fft` for definitions and conventions used.

    See `rfft` for definitions and conventions used for real input.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.zeros((3, 2, 2))
    >>> a[0, 0, 0] = 3 * 2 * 2
    >>> mt.fft.irfftn(a).execute()
    array([[[ 1.,  1.],
            [ 1.,  1.]],
           [[ 1.,  1.],
            [ 1.,  1.]],
           [[ 1.,  1.],
            [ 1.,  1.]]])

    """
    a = astensor(a)
    axes = validate_fftn(a, s=s, axes=axes, norm=norm)
    op = TensorIRFFTN(shape=s, axes=axes, norm=norm, dtype=np.dtype(np.float_))
    return op(a)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Number

from ...tensor import tensor as astensor
from ...tensor.ufunc.ufunc import UFUNC_TO_TENSOR_FUNC
from ..core import DATAFRAME_TYPE, SERIES_TYPE


def _check_arg(arg):
    if isinstance(arg, Number):
        return True

    if isinstance(arg, (DATAFRAME_TYPE, SERIES_TYPE)):
        return True

    try:
        astensor(arg)
        return True
    except ValueError:
        return False


def _array_ufunc(_, ufunc, method, *inputs, **kwargs):
    out = kwargs.get('out', tuple())
    for x in inputs + out:
        if not _check_arg(x):
            return NotImplemented

    if method == '__call__':
        if ufunc.signature is not None:
            return NotImplemented
        if ufunc not in UFUNC_TO_TENSOR_FUNC:
            return NotImplemented

        # we delegate numpy ufunc to tensor ufunc,
        # tensor ufunc will handle Mars DataFrame properly.
        tensor_func = UFUNC_TO_TENSOR_FUNC[ufunc]
        return tensor_func(*inputs, **kwargs)

    return NotImplemented



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gzip import GzipFile
from typing import BinaryIO

try:
    import lz4
    import lz4.frame
except ImportError:  # pragma: no cover
    lz4 = None


_compressions = {
    'gzip': lambda f: GzipFile(fileobj=f)
}

if lz4:
    _compressions['lz4'] = lz4.frame.open


def compress(file: BinaryIO,
             compress_type: str) -> BinaryIO:
    """
    Return a compressed file object.

    Parameters
    ----------
    file:
        file object.
    compress_type: str
       compression type.

    Returns
    -------
    compressed_file:
        compressed file object.
    """
    try:
        compress_ = _compressions[compress_type]
    except KeyError:  # pragma: no cover
        raise ValueError(f'Unknown compress type: {compress_type}, '
                         f'available include: {", ".join(_compressions)}')

    return compress_(file)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os
import json
from collections import defaultdict

import numpy as np

from ....context import get_context, RunningMode
from .... import opcodes as OperandDef
from ....remote.run_script import RunScript
from ....serialize import BytesField, Int32Field, DictField, StringField
from ....utils import to_binary
from ..utils import pick_workers


class RunTensorFlow(RunScript):
    _op_type_ = OperandDef.RUN_TENSORFLOW

    _code = BytesField('code')
    _n_workers = Int32Field('n_workers')
    _n_ps = Int32Field('n_ps')
    _tf_config = DictField('tf_config')
    _port = Int32Field('port')
    # used for chunk op
    _tf_task_type = StringField('tf_task_type')
    _tf_task_index = Int32Field('tf_task_index')

    def __init__(self, n_workers=None, n_ps=None, tf_config=None, port=None,
                 tf_task_type=None, tf_task_index=None, gpu=None, **kw):
        super().__init__(mode='spawn', _n_workers=n_workers, _n_ps=n_ps,
                         _tf_config=tf_config, _port=port, _tf_task_type=tf_task_type,
                         _tf_task_index=tf_task_index, _gpu=gpu, **kw)

    @property
    def n_workers(self):
        return self._n_workers

    @property
    def n_ps(self):
        return self._n_ps or 0

    @property
    def n_roles(self):
        return self._n_workers + self._n_ps

    @property
    def tf_config(self):
        return self._tf_config

    @property
    def port(self):
        return self._port

    @property
    def tf_task_type(self):
        return self._tf_task_type

    @property
    def tf_task_index(self):
        return self._tf_task_index

    def __call__(self):
        return self.new_tileable(None)

    @classmethod
    def tile(cls, op):
        ctx = get_context()

        port = op.port or 2221
        cluster_conf = {'worker': []}
        if op.n_ps > 0:
            cluster_conf['ps'] = []
        n_workers = op.n_workers

        out_chunks = []
        if ctx.running_mode != RunningMode.distributed:
            worker_addr = '127.0.0.1'
            port_iter = itertools.count(port)

            for i in range(op.n_roles):
                chunk_op = op.copy().reset_key()
                addr = f'{worker_addr}:{next(port_iter)}'
                chunk_op._tf_task_type = tp = 'worker' if i < n_workers else 'ps'
                chunk_op._tf_task_index = idx = i if i < n_workers else i - n_workers
                cluster_conf[tp].append(addr)
                chunk_op._tf_config = {'cluster': cluster_conf,
                                       'task': {'type': tp, 'index': idx}}
                out_chunks.append(chunk_op.new_chunk(None, index=(i,)))
        else:
            worker_addresses = ctx.get_worker_addresses()
            picked_workers = pick_workers(worker_addresses, op.n_roles)
            worker_to_port_iter = defaultdict(lambda: itertools.count(port))

            for i, worker in enumerate(picked_workers):
                worker_addr = worker.rsplit(':', 1)[0]
                chunk_op = op.copy().reset_key()
                addr = f'{worker_addr}:{next(worker_to_port_iter[worker_addr])}'
                # tell graph actor that the chunk should be executed on the exact worker
                chunk_op._expect_worker = worker
                tp = 'worker' if i < n_workers else 'ps'
                chunk_op._tf_task_type = tp
                idx = i if i < n_workers else i - n_workers
                chunk_op._tf_task_index = idx
                cluster_conf[tp].append(addr)
                chunk_op._tf_config = {'cluster': cluster_conf,
                                       'task': {'type': tp, 'index': idx}}
                out_chunks.append(chunk_op.new_chunk(None, index=(i,)))

        new_op = op.copy()
        return new_op.new_tileables(op.inputs, chunks=out_chunks,
                                    nsplits=(tuple(np.nan for _ in range(len(out_chunks))),))

    @classmethod
    def _build_envs(cls, ctx, op):
        envs = super()._build_envs(ctx, op)
        envs.update({'TF_CONFIG': json.dumps(op.tf_config)})
        return envs

    @classmethod
    def execute(cls, ctx, op):
        if op.merge:
            return super().execute(ctx, op)

        assert ctx.get_local_address() == op.expect_worker

        super().execute(ctx, op)

        if op.tf_task_type == 'worker' and op.tf_task_index == 0:
            ctx[op.outputs[0].key] = {'status': 'ok'}
        else:
            ctx[op.outputs[0].key] = {}


def run_tensorflow_script(script, n_workers, n_ps=0, gpu=None, command_argv=None,
                          retry_when_fail=False, session=None, run_kwargs=None, port=None):
    """
    Run TensorFlow script in Mars cluster.

    :param script: script to run
    :type script: str or file-like object
    :param n_workers: number of TensorFlow workers
    :param n_ps: number of TensorFlow ps, optional
    :param gpu: run TensorFlow script on GPU
    :param command_argv: extra command args for script
    :param retry_when_fail: bool, default False. If True, retry when function failed.
    :param session: Mars session, if not provided, will use default one
    :param run_kwargs: extra kwargs for session.run
    :param port: port of TensorFlow worker or ps, will automatically increase for the same worker
    :return: return {'status': 'ok'} if succeeded, or error raised
    """
    if int(n_workers) <= 0:
        raise ValueError('n_workers should be at least 1')
    if int(n_ps) < 0:
        raise ValueError('n_ps should be at least 0')
    if hasattr(script, 'read'):
        code = script.read()
    else:
        with open(os.path.abspath(script), 'rb') as f:
            code = f.read()

    op = RunTensorFlow(code=to_binary(code), n_workers=int(n_workers), n_ps=int(n_ps),
                       retry_when_fail=retry_when_fail, gpu=gpu,
                       port=port, command_args=command_argv)
    return op().execute(session=session, **(run_kwargs or {})).fetch(session=session)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import logging
from enum import Enum

from ...actors import ActorNotExist
from ...errors import WorkerDead
from ...utils import classproperty

logger = logging.getLogger(__name__)


class OperandState(Enum):
    __order__ = 'UNSCHEDULED READY RUNNING FINISHED CACHED FREED FATAL CANCELLING CANCELLED'

    UNSCHEDULED = 'unscheduled'
    READY = 'ready'
    RUNNING = 'running'
    FINISHED = 'finished'
    CACHED = 'cached'
    FREED = 'freed'
    FATAL = 'fatal'
    CANCELLING = 'cancelling'
    CANCELLED = 'cancelled'

    @classproperty
    def STORED_STATES(self):
        """
        States on which the data of the operand is stored
        """
        return self.FINISHED, self.CACHED

    @classproperty
    def SUCCESSFUL_STATES(self):
        """
        States on which the operand is executed successfully
        """
        return self.FINISHED, self.CACHED, self.FREED

    @classproperty
    def TERMINATED_STATES(self):
        """
        States on which the operand has already terminated
        """
        return self.FINISHED, self.CACHED, self.FREED, self.FATAL, self.CANCELLED


@contextlib.contextmanager
def rewrite_worker_errors(ignore_error=False):
    rewrite = False
    try:
        yield
    except (BrokenPipeError, ConnectionRefusedError, ActorNotExist):
        # we don't raise here, as we do not want
        # the actual stack be dumped
        logger.exception('Worker connection error detected, will raise WorkerDead')
        rewrite = not ignore_error
    if rewrite:
        raise WorkerDead


_op_cls_to_actor = dict()


def get_operand_actor_class(op_cls):
    try:
        return _op_cls_to_actor[op_cls]
    except KeyError:
        for super_cls in op_cls.__mro__:
            try:
                actor_cls = _op_cls_to_actor[op_cls] = _op_cls_to_actor[super_cls]
                return actor_cls
            except KeyError:
                continue
        raise KeyError(f'Operand type {op_cls} not supported')  # pragma: no cover


def register_operand_class(op_cls, actor_cls):
    _op_cls_to_actor[op_cls] = actor_cls



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

from ..serialize.dataserializer import SerialType, CompressType, get_compressobj, get_decompressobj, \
    HEADER_LENGTH, file_header, read_file_header, write_file_header, SERIAL_VERSION


class WorkerBufferIO(object):
    """
    File-like object handling data with file header
    """
    def __init__(self, mode='r', compress_in=None, compress_out=None, block_size=8192):
        """
        :param mode: 'r' indicates read, or 'w' indicates write
        :param compress_in: compression type inside
        :param compress_out: compression type outside
        :param block_size: size of data block when copying
        """
        self._mode = mode
        self._block_size = block_size
        self._compress_type_in = compress_in or CompressType.NONE
        self._compress_type_out = compress_out or CompressType.NONE

        if 'w' in mode:
            self._compressor_in = None
            self._decompressor_out = None
        else:
            self._remain_buf = None
            self._remain_offset = None
            self._block_iterator = self._iter_blocks()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def _read_block(self, size):
        """
        Read a data block from data source with given size.
        """
        raise NotImplementedError

    def _read_header(self):
        """
        Read the data header of data source.
        """
        raise NotImplementedError

    def _iter_blocks(self):
        """
        Returns a generator providing data blocks. The sizes of data blocks can vary.
        """
        bio = BytesIO()
        compressor_out = get_compressobj(self._compress_type_out)

        header = self._read_header()
        self._compress_type_in = header.compress
        decompressor_in = get_decompressobj(header.compress)

        new_header = file_header(header.type, header.version, header.nbytes, self._compress_type_out)
        write_file_header(bio, new_header)

        if self._compress_type_in == self._compress_type_out:
            handle_block = lambda v: v

            compressor_out = get_compressobj(CompressType.NONE)
            decompressor_in = get_decompressobj(CompressType.NONE)
        else:
            handle_block = lambda v: compressor_out.compress(decompressor_in.decompress(v))
            if hasattr(compressor_out, 'begin'):
                bio.write(compressor_out.begin())

        # yield file header and compress header
        yield bio.getvalue()

        copy_size = self._block_size
        while True:
            block = self._read_block(copy_size)
            if not block:
                break
            buf = handle_block(block)
            if buf:
                # only yield when some data are produced by the compressor
                yield buf

        if hasattr(compressor_out, 'flush'):
            yield compressor_out.flush()

    def read(self, size=-1):
        bio = BytesIO()
        if self._remain_buf is not None:
            if size < 0 or len(self._remain_buf) <= self._remain_offset + size:
                bio.write(self._remain_buf[self._remain_offset:])
                self._remain_buf = None
            else:
                right = self._remain_offset + size
                bio.write(self._remain_buf[self._remain_offset:right])
                self._remain_offset = right
            if bio.tell() == size:
                return bio.getvalue()
        while True:
            try:
                block = next(self._block_iterator)
            except StopIteration:
                break
            if size < 0 or bio.tell() + len(block) <= size:
                bio.write(block)
                self._remain_buf = None
            else:
                offset = self._remain_offset = size - bio.tell()
                self._remain_buf = buf = memoryview(block)
                bio.write(buf[:offset])
            if bio.tell() == size:
                break
        return bio.getvalue()

    def _write_header(self, header):
        """
        Write a data header to the output file.
        """
        raise NotImplementedError

    def _write_block(self, d):
        """
        Write a data block to the output file.
        """
        raise NotImplementedError

    def _write_with_compression(self, mv):
        data_len = len(mv)
        copy_size = self._block_size
        offset = 0

        if self._compressor_in is not None:
            compressor_in = self._compressor_in
        else:
            compressor_in = self._compressor_in = get_compressobj(self._compress_type_in)
            if hasattr(compressor_in, 'begin') and self._compress_type_in != self._compress_type_out:
                self._write_block(compressor_in.begin())

        if self._compress_type_in == self._compress_type_out:
            self._compressor_in = get_compressobj(CompressType.NONE)
            self._decompressor_out = get_decompressobj(CompressType.NONE)

            handle_block = lambda v: v
        else:
            handle_block = lambda v: compressor_in.compress(decompressor_out.decompress(v))

        decompressor_out = self._decompressor_out

        while offset < data_len:
            right = min(data_len, offset + copy_size)
            cblock = handle_block(mv[offset:right])
            if cblock:
                self._write_block(cblock)
            offset = right

    def write(self, d):
        size = len(d)
        mv = memoryview(d)
        if self._decompressor_out is not None:
            # header already processed, we can continue
            # with decompression
            self._write_with_compression(mv)
        else:
            # header not processed, we need to read header
            # to get compression method and build decompressor
            if size < HEADER_LENGTH:
                raise IOError('Block size too small')

            header = read_file_header(mv)
            self._compress_type_out = header.compress
            new_header = file_header(header.type, header.version, header.nbytes, self._compress_type_in)
            self._write_header(new_header)

            self._decompressor_out = get_decompressobj(header.compress)
            if size > HEADER_LENGTH:
                self._write_with_compression(mv[HEADER_LENGTH:])

    def close(self):
        if 'w' in self._mode and hasattr(self._compressor_in, 'flush'):
            self._write_block(self._compressor_in.flush())

        self._compressor_in = self._decompressor_out = None
        self._remain_buf = self._block_iterator = None


class ArrowBufferIO(WorkerBufferIO):
    """
    File-like object mocking object stored in shared memory as file with header
    """
    def __init__(self, buf, mode='r', compress_in=None, compress_out=None, block_size=8192):
        super().__init__(
            mode=mode, compress_in=compress_in, compress_out=compress_out, block_size=block_size)

        self._buf = buf
        if 'r' in mode:
            self._mv = memoryview(buf)
            self._nbytes = len(buf)
            self._buf_offset = 0
            self._writer = None
        else:
            import pyarrow
            self._writer = pyarrow.FixedSizeBufferWriter(buf)
            self._writer.set_memcopy_threads(6)

    def _read_header(self):
        return file_header(SerialType.ARROW, SERIAL_VERSION, self._nbytes, self._compress_type_in)

    def _read_block(self, size):
        right = min(self._nbytes, self._buf_offset + size)
        ret = self._mv[self._buf_offset:right]
        self._buf_offset = right
        return ret

    def _write_header(self, header):
        pass

    def _write_block(self, d):
        self._writer.write(d)

    def close(self):
        if self._writer is not None:
            self._writer.close()
        self._writer = self._mv = self._buf = None

        super().close()


class FileBufferIO(WorkerBufferIO):
    """
    File-like object handling input of one compression type and outputs another
    """
    def __init__(self, file, mode='r', compress_in=None, compress_out=None,
                 block_size=8192, managed=True, header=None, file_end=None):
        super().__init__(
            mode=mode, compress_in=compress_in, compress_out=compress_out, block_size=block_size)

        self._managed = managed
        self._file = file
        self._header = header
        self._file_end = file_end

    def _read_header(self):
        if self._header is not None:
            return self._header
        return read_file_header(self._file)

    def _read_block(self, size):
        if self._file_end is not None:
            max_size = self._file_end - self._file.tell()
            if size == -1:
                size = max_size
            size = max(min(size, max_size), 0)
        return self._file.read(size)

    def _write_header(self, header):
        return write_file_header(self._file, header)

    def _write_block(self, d):
        self._file.write(d)

    def close(self):
        super().close()
        if self._file:
            if self._managed:
                self._file.close()
            else:
                self._file.flush()
        self._file = None



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import OrderedDict
from distutils.version import LooseVersion

import pandas as pd

from ....serialize import Int64Field, BoolField, Int32Field, Float64Field
from ...utils import validate_axis
from ..core import Window

_default_min_period_1 = LooseVersion(pd.__version__) >= '1.1.0'
_pd_1_3_repr = LooseVersion(pd.__version__) >= '1.3.0'


class EWM(Window):
    _alpha = Float64Field('alpha')
    _min_periods = Int64Field('min_periods')
    _adjust = BoolField('adjust')
    _ignore_na = BoolField('ignore_na')
    _axis = Int32Field('axis')

    def __init__(self, alpha=None, min_periods=None, adjust=None, ignore_na=None, axis=None, **kw):
        super().__init__(_alpha=alpha, _min_periods=min_periods, _adjust=adjust, _ignore_na=ignore_na,
                         _axis=axis, **kw)

    @property
    def alpha(self):
        return self._alpha

    @property
    def min_periods(self):
        return self._min_periods

    @property
    def adjust(self):
        return self._adjust

    @property
    def ignore_na(self):
        return self._ignore_na

    @property
    def axis(self):
        return self._axis

    @property
    def params(self):
        p = OrderedDict()
        for k in ['alpha', 'min_periods', "adjust", "ignore_na", "axis"]:
            p[k] = getattr(self, k)
        return p

    def __call__(self, df):
        return df.ewm(**self.params)

    def _repr(self, params):
        com = 1.0 / params.pop('alpha') - 1
        params['com'] = int(com) if _pd_1_3_repr and com == math.floor(com) else com
        try:
            params.move_to_end('com', last=False)
        except AttributeError:  # pragma: no cover
            pass
        return super()._repr(params)

    def _repr_name(self):
        try:
            from pandas.core.window import ExponentialMovingWindow  # noqa: F401
            return 'ExponentialMovingWindow'
        except ImportError:  # pragma: no cover
            return 'EWM'

    def aggregate(self, func):
        from .aggregation import DataFrameEwmAgg

        params = self.params
        params['alpha_ignore_na'] = params.pop('ignore_na', False)
        params['validate_columns'] = LooseVersion(pd.__version__) < '1.0.0'
        op = DataFrameEwmAgg(func=func, **params)
        return op(self)

    agg = aggregate

    def mean(self):
        return self.aggregate('mean')

    def var(self):
        return self.aggregate('var')

    def std(self):
        return self.aggregate('std')


def ewm(obj, com=None, span=None, halflife=None, alpha=None, min_periods=0, adjust=True,
        ignore_na=False, axis=0):
    r"""
    Provide exponential weighted functions.

    Parameters
    ----------
    com : float, optional
        Specify decay in terms of center of mass,
        :math:`\alpha = 1 / (1 + com),\text{ for } com \geq 0`.
    span : float, optional
        Specify decay in terms of span,
        :math:`\alpha = 2 / (span + 1),\text{ for } span \geq 1`.
    halflife : float, optional
        Specify decay in terms of half-life,
        :math:`\alpha = 1 - exp(log(0.5) / halflife),\text{for} halflife > 0`.
    alpha : float, optional
        Specify smoothing factor :math:`\alpha` directly,
        :math:`0 < \alpha \leq 1`.
    min_periods : int, default 0
        Minimum number of observations in window required to have a value
        (otherwise result is NA).
    adjust : bool, default True
        Divide by decaying adjustment factor in beginning periods to account
        for imbalance in relative weightings
        (viewing EWMA as a moving average).
    ignore_na : bool, default False
        Ignore missing values when calculating weights;
        specify True to reproduce pre-0.15.0 behavior.
    axis : {0 or 'index', 1 or 'columns'}, default 0
        The axis to use. The value 0 identifies the rows, and 1
        identifies the columns.

    Returns
    -------
    DataFrame
        A Window sub-classed for the particular operation.

    See Also
    --------
    rolling : Provides rolling window calculations.
    expanding : Provides expanding transformations.

    Notes
    -----
    Exactly one of center of mass, span, half-life, and alpha must be provided.

    Allowed values and relationship between the parameters are specified in the
    parameter descriptions above; see the link at the end of this section for
    a detailed explanation.

    When adjust is True (default), weighted averages are calculated using
    weights (1-alpha)**(n-1), (1-alpha)**(n-2), ..., 1-alpha, 1.

    When adjust is False, weighted averages are calculated recursively as:

       weighted_average[0] = arg[0];
       weighted_average[i] = (1-alpha)*weighted_average[i-1] + alpha*arg[i].

    When ignore_na is False (default), weights are based on absolute positions.
    For example, the weights of x and y used in calculating the final weighted
    average of [x, None, y] are (1-alpha)**2 and 1 (if adjust is True), and
    (1-alpha)**2 and alpha (if adjust is False).

    When ignore_na is True (reproducing pre-0.15.0 behavior), weights are based
    on relative positions. For example, the weights of x and y used in
    calculating the final weighted average of [x, None, y] are 1-alpha and 1
    (if adjust is True), and 1-alpha and alpha (if adjust is False).

    More details can be found at
    https://pandas.pydata.org/pandas-docs/stable/user_guide/computation.html#exponentially-weighted-windows

    Examples
    --------
    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df.execute()
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0
    >>> df.ewm(com=0.5).mean().execute()
              B
    0  0.000000
    1  0.750000
    2  1.615385
    3  1.615385
    4  3.670213
    """
    axis = validate_axis(axis, obj)

    decay_count = 0
    for arg in (com, span, halflife, alpha):
        if arg is not None:
            decay_count += 1

    if decay_count == 0:
        raise ValueError('Must pass one of comass, span, halflife, or alpha')
    if decay_count > 1:
        raise ValueError('comass, span, halflife, and alpha are mutually exclusive')

    if com is not None:
        if com < 0:
            raise ValueError('comass must satisfy: comass >= 0')
        alpha = 1.0 / (1 + com)
    elif span is not None:
        if span < 1:
            raise ValueError('span must satisfy: span >= 1')
        alpha = 2.0 / (1 + span)
    elif halflife is not None:
        if halflife <= 0:
            raise ValueError('halflife must satisfy: halflife > 0')
        alpha = 1.0 - math.exp(math.log(0.5) / halflife)
    if alpha <= 0 or alpha > 1:
        raise ValueError('alpha must satisfy: 0 < alpha <= 1')

    if not adjust and not ignore_na:
        raise NotImplementedError('adjust == False when ignore_na == False not implemented')
    if axis == 1:
        raise NotImplementedError('axis other than 0 is not supported')

    if alpha == 1:
        return obj.expanding(min_periods=min_periods, axis=axis)

    if _default_min_period_1:
        min_periods = min_periods or 1

    return EWM(input=obj, alpha=alpha, min_periods=min_periods, adjust=adjust,
               ignore_na=ignore_na, axis=axis)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import scipy

    from .err_fresnel import erf, TensorErf
    from .gamma_funcs import (
        gamma, TensorGamma,
        gammaln, TensorGammaln,
        loggamma, TensorLogGamma,
        gammasgn, TensorGammaSgn,
        gammainc, TensorGammaInc,
        gammaincinv, TensorGammaIncInv,
        gammaincc, TensorGammaIncc,
        gammainccinv, TensorGammaInccInv,
        beta, TensorBeta,
        betaln, TensorBetaLn,
        betainc, TensorBetaInc,
        betaincinv, TensorBetaIncInv,
        psi, TensorPsi,
        rgamma, TensorRGamma,
        polygamma, TensorPolyGamma,
        multigammaln, TensorMultiGammaLn,
        digamma, TensorDiGamma,
        poch, TensorPoch,
    )
    from .info_theory import (
        entr, TensorEntr,
        rel_entr, TensorRelEntr,
        kl_div, TensorKlDiv,
    )
    from .convenience import (
        xlogy, TensorXLogY,
    )
except ImportError:  # pragma: no cover
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorReduction, TensorReductionMixin


class TensorProd(TensorReduction, TensorReductionMixin):
    _op_type_ = OperandDef.PROD
    _func_name = 'prod'

    def __init__(self, axis=None, dtype=None, keepdims=None, combine_size=None, stage=None, **kw):
        stage = self._rewrite_stage(stage)
        super().__init__(_axis=axis, _dtype=dtype, _keepdims=keepdims,
                         _combine_size=combine_size, _stage=stage, **kw)


def prod(a, axis=None, dtype=None, out=None, keepdims=None, combine_size=None):
    """
    Return the product of tensor elements over a given axis.

    Parameters
    ----------
    a : array_like
        Input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which a product is performed.  The default,
        axis=None, will calculate the product of all the elements in the
        input tensor. If axis is negative it counts from the last to the
        first axis.

        If axis is a tuple of ints, a product is performed on all of the
        axes specified in the tuple instead of a single axis or all the
        axes as before.
    dtype : dtype, optional
        The type of the returned tensor, as well as of the accumulator in
        which the elements are multiplied.  The dtype of `a` is used by
        default unless `a` has an integer dtype of less precision than the
        default platform integer.  In that case, if `a` is signed then the
        platform integer is used while if `a` is unsigned then an unsigned
        integer of the same precision as the platform integer is used.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must have
        the same shape as the expected output, but the type of the output
        values will be cast if necessary.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the
        result as dimensions with size one. With this option, the result
        will broadcast correctly against the input array.

        If the default value is passed, then `keepdims` will not be
        passed through to the `prod` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    product_along_axis : Tensor, see `dtype` parameter above.
        An tensor shaped as `a` but with the specified axis removed.
        Returns a reference to `out` if specified.

    See Also
    --------
    Tensor.prod : equivalent method

    Notes
    -----
    Arithmetic is modular when using integer types, and no error is
    raised on overflow.  That means that, on a 32-bit platform:

    >>> import mars.tensor as mt

    >>> x = mt.array([536870910, 536870910, 536870910, 536870910])
    >>> mt.prod(x).execute()  # random
    16

    The product of an empty array is the neutral element 1:

    >>> mt.prod([]).execute()
    1.0

    Examples
    --------
    By default, calculate the product of all elements:

    >>> mt.prod([1.,2.]).execute()
    2.0

    Even when the input array is two-dimensional:

    >>> mt.prod([[1.,2.],[3.,4.]]).execute()
    24.0

    But we can also specify the axis over which to multiply:

    >>> mt.prod([[1.,2.],[3.,4.]], axis=1).execute()
    array([  2.,  12.])

    If the type of `x` is unsigned, then the output type is
    the unsigned platform integer:

    >>> x = mt.array([1, 2, 3], dtype=mt.uint8)
    >>> mt.prod(x).dtype == mt.uint
    True

    If `x` is of a signed integer type, then the output type
    is the default platform integer:

    >>> x = mt.array([1, 2, 3], dtype=mt.int8)
    >>> mt.prod(x).dtype == int
    True

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.empty((1,), dtype=a.dtype).prod().dtype
    op = TensorProd(axis=axis, dtype=dtype, keepdims=keepdims, combine_size=combine_size)
    return op(a, out=out)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
import operator
import pickle
from collections import defaultdict
from functools import reduce

import numpy as np

from .... import opcodes
from ....context import get_context, RunningMode
from ....core import ExecutableTuple
from ....operands import MergeDictOperand, OutputType
from ....serialize import DictField, Int32Field, KeyField, ListField, StringField, ValueType
from ...utils import concat_chunks, collect_ports
from ._align import align_data_set
from .core import LGBMModelType, get_model_cls_from_type

logger = logging.getLogger(__name__)


class LGBMTrain(MergeDictOperand):
    _op_type_ = opcodes.LGBM_TRAIN

    _model_type = Int32Field('model_type', on_serialize=lambda x: x.value,
                             on_deserialize=LGBMModelType)
    _params = DictField('params', key_type=ValueType.string)
    _data = KeyField('data')
    _label = KeyField('label')
    _sample_weight = KeyField('sample_weight')
    _init_score = KeyField('init_score')
    _kwds = DictField('kwds', key_type=ValueType.string)

    _eval_datas = ListField('eval_datas', ValueType.key)
    _eval_labels = ListField('eval_labels', ValueType.key)
    _eval_sample_weights = ListField('eval_sample_weights', ValueType.key)
    _eval_init_scores = ListField('eval_init_scores', ValueType.key)

    _workers = ListField('workers', ValueType.string)
    _worker_id = Int32Field('worker_id')
    _worker_ports = KeyField('worker_ports')

    _tree_learner = StringField('tree_learner')
    _timeout = Int32Field('timeout')

    def __init__(self, model_type=None, data=None, label=None, sample_weight=None, init_score=None,
                 eval_datas=None, eval_labels=None, eval_sample_weights=None, eval_init_scores=None,
                 params=None, kwds=None, workers=None, worker_id=None, worker_ports=None,
                 tree_learner=None, timeout=None, **kw):
        super().__init__(_model_type=model_type, _params=params, _data=data, _label=label,
                         _sample_weight=sample_weight, _init_score=init_score, _eval_datas=eval_datas,
                         _eval_labels=eval_labels, _eval_sample_weights=eval_sample_weights,
                         _eval_init_scores=eval_init_scores, _kwds=kwds, _workers=workers,
                         _worker_id=worker_id, _worker_ports=worker_ports, _tree_learner=tree_learner,
                         _timeout=timeout, **kw)
        if self.output_types is None:
            self.output_types = [OutputType.object]

    @property
    def model_type(self) -> LGBMModelType:
        return self._model_type

    @property
    def data(self):
        return self._data

    @property
    def label(self):
        return self._label

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def init_score(self):
        return self._init_score

    @property
    def eval_datas(self) -> list:
        return self._eval_datas or []

    @property
    def eval_labels(self) -> list:
        return self._eval_labels or []

    @property
    def eval_sample_weights(self) -> list:
        return self._eval_sample_weights or []

    @property
    def eval_init_scores(self) -> list:
        return self._eval_init_scores or []

    @property
    def params(self) -> dict:
        return self._params or dict()

    @property
    def kwds(self) -> dict:
        return self._kwds or dict()

    @property
    def workers(self) -> list:
        return self._workers

    @property
    def worker_id(self) -> int:
        return self._worker_id

    @property
    def worker_ports(self):
        return self._worker_ports

    @property
    def timeout(self) -> int:
        return self._timeout

    @property
    def tree_learner(self) -> str:
        return self._tree_learner

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        it = iter(inputs)
        for attr in ['_data', '_label', '_sample_weight', '_init_score']:
            if getattr(self, attr) is not None:
                setattr(self, attr, next(it))
        for attr in ['_eval_datas', '_eval_labels', '_eval_sample_weights', '_eval_init_scores']:
            new_list = []
            for c in getattr(self, attr, None) or []:
                if c is not None:
                    new_list.append(next(it))
            setattr(self, attr, new_list or None)

        if self._worker_ports is not None:
            self._worker_ports = next(it)

    def __call__(self):
        inputs = []
        for attr in ['_data', '_label', '_sample_weight', '_init_score']:
            if getattr(self, attr) is not None:
                inputs.append(getattr(self, attr))
        for attr in ['_eval_datas', '_eval_labels', '_eval_sample_weights', '_eval_init_scores']:
            for c in getattr(self, attr, None) or []:
                if c is not None:
                    inputs.append(c)
        return self.new_tileable(inputs)

    @staticmethod
    def _get_data_chunks_workers(ctx, data):
        # data_chunk.inputs is concat, and concat's input is the co-allocated chunks
        metas = ctx.get_chunk_metas([c.key for c in data.chunks])
        return [m.workers[0] for m in metas]

    @staticmethod
    def _concat_chunks_by_worker(chunks, chunk_workers):
        worker_to_chunks = defaultdict(list)
        for chunk, worker in zip(chunks, chunk_workers):
            worker_to_chunks[worker].append(chunk)
        worker_to_concat = dict()
        for worker, chunks in worker_to_chunks.items():
            worker_to_concat[worker] = concat_chunks(chunks)
        return worker_to_concat

    @classmethod
    def tile(cls, op: "LGBMTrain"):
        ctx = get_context()
        if ctx.running_mode != RunningMode.distributed:
            assert all(len(inp.chunks) == 1 for inp in op.inputs)

            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk([inp.chunks[0] for inp in op.inputs],
                                           shape=(1,), index=(0,))
            new_op = op.copy()
            return new_op.new_tileables(op.inputs, chunks=[out_chunk], nsplits=((1,),))
        else:
            data = op.data
            worker_to_args = defaultdict(dict)

            workers = cls._get_data_chunks_workers(ctx, data)

            for arg in ['_data', '_label', '_sample_weight', '_init_score']:
                if getattr(op, arg) is not None:
                    for worker, chunk in cls._concat_chunks_by_worker(
                            getattr(op, arg).chunks, workers).items():
                        worker_to_args[worker][arg] = chunk

            if op.eval_datas:
                eval_workers_list = [cls._get_data_chunks_workers(ctx, d) for d in op.eval_datas]
                extra_workers = reduce(operator.or_, (set(w) for w in eval_workers_list)) - set(workers)
                worker_remap = dict(zip(extra_workers, itertools.cycle(workers)))
                if worker_remap:
                    eval_workers_list = [[worker_remap.get(w, w) for w in wl] for wl in eval_workers_list]

                for arg in ['_eval_datas', '_eval_labels', '_eval_sample_weights', '_eval_init_scores']:
                    if getattr(op, arg):
                        for tileable, eval_workers in zip(getattr(op, arg), eval_workers_list):
                            for worker, chunk in cls._concat_chunks_by_worker(
                                    tileable.chunks, eval_workers).items():
                                if arg not in worker_to_args[worker]:
                                    worker_to_args[worker][arg] = []
                                worker_to_args[worker][arg].append(chunk)

            out_chunks = []
            workers = list(set(workers))
            for worker_id, worker in enumerate(workers):
                chunk_op = op.copy().reset_key()
                chunk_op._expect_worker = worker

                input_chunks = []
                concat_args = worker_to_args.get(worker, {})
                for arg in ['_data', '_label', '_sample_weight', '_init_score',
                            '_eval_datas', '_eval_labels', '_eval_sample_weights', '_eval_init_scores']:
                    arg_val = getattr(op, arg)
                    if arg_val:
                        arg_chunk = concat_args.get(arg)
                        setattr(chunk_op, arg, arg_chunk)
                        if isinstance(arg_chunk, list):
                            input_chunks.extend(arg_chunk)
                        else:
                            input_chunks.append(arg_chunk)

                worker_ports_chunk = collect_ports(workers, op.data)._inplace_tile().chunks[0]
                input_chunks.append(worker_ports_chunk)

                chunk_op._workers = workers
                chunk_op._worker_ports = worker_ports_chunk
                chunk_op._worker_id = worker_id

                data_chunk = concat_args['_data']
                out_chunk = chunk_op.new_chunk(input_chunks, shape=(np.nan,), index=data_chunk.index[:1])
                out_chunks.append(out_chunk)

            new_op = op.copy()
            return new_op.new_tileables(op.inputs, chunks=out_chunks,
                                        nsplits=((np.nan for _ in out_chunks),))

    @classmethod
    def execute(cls, ctx, op: "LGBMTrain"):
        if op.merge:
            return super().execute(ctx, op)

        from lightgbm.basic import _safe_call, _LIB

        data_val = ctx[op.data.key]
        data_val = data_val.spmatrix if hasattr(data_val, 'spmatrix') else data_val

        label_val = ctx[op.label.key]
        sample_weight_val = ctx[op.sample_weight.key] if op.sample_weight is not None else None
        init_score_val = ctx[op.init_score.key] if op.init_score is not None else None

        if op.eval_datas is None:
            eval_set, eval_sample_weight, eval_init_score = None, None, None
        else:
            eval_set, eval_sample_weight, eval_init_score = [], [], []
            for data, label in zip(op.eval_datas, op.eval_labels):
                data_eval = ctx[data.key]
                data_eval = data_eval.spmatrix if hasattr(data_eval, 'spmatrix') else data_eval
                eval_set.append((data_eval, ctx[label.key]))
            for weight in op.eval_sample_weights:
                eval_sample_weight.append(ctx[weight.key] if weight is not None else None)
            for score in op.eval_init_scores:
                eval_init_score.append(ctx[score.key] if score is not None else None)

            eval_set = eval_set or None
            eval_sample_weight = eval_sample_weight or None
            eval_init_score = eval_init_score or None

        params = op.params.copy()
        # if model is trained, remove unsupported parameters
        params.pop('out_dtype_', None)
        if ctx.running_mode == RunningMode.distributed:
            worker_ports = ctx[op.worker_ports.key]
            worker_ips = [worker.split(':', 1)[0] for worker in op.workers]
            worker_endpoints = [f'{worker}:{port}' for worker, port in zip(worker_ips, worker_ports)]

            params['machines'] = ','.join(worker_endpoints)
            params['time_out'] = op.timeout
            params['num_machines'] = len(worker_endpoints)
            params['local_listen_port'] = worker_ports[op.worker_id]

            if (op.tree_learner or '').lower() not in {'data', 'feature', 'voting'}:
                logger.warning('Parameter tree_learner not set or set to incorrect value '
                               f'{op.tree_learner}, using "data" as default')
                params['tree_learner'] = 'data'
            else:
                params['tree_learner'] = op.tree_learner

        try:
            model_cls = get_model_cls_from_type(op.model_type)
            model = model_cls(**params)
            model.fit(data_val, label_val, sample_weight=sample_weight_val, init_score=init_score_val,
                      eval_set=eval_set, eval_sample_weight=eval_sample_weight,
                      eval_init_score=eval_init_score, **op.kwds)

            if op.model_type == LGBMModelType.RANKER or \
                    op.model_type == LGBMModelType.REGRESSOR:
                model.set_params(out_dtype_=np.dtype('float'))
            elif hasattr(label_val, 'dtype'):
                model.set_params(out_dtype_=label_val.dtype)
            else:
                model.set_params(out_dtype_=label_val.dtypes[0])

            ctx[op.outputs[0].key] = pickle.dumps(model)
        finally:
            _safe_call(_LIB.LGBM_NetworkFree())


def train(params, train_set, eval_sets=None, **kwargs):
    eval_sets = eval_sets or []
    model_type = kwargs.pop('model_type', LGBMModelType.CLASSIFIER)

    evals_result = kwargs.pop('evals_result', dict())
    session = kwargs.pop('session', None)
    run_kwargs = kwargs.pop('run_kwargs', None)
    if run_kwargs is None:
        run_kwargs = dict()
    timeout = kwargs.pop('timeout', 120)
    base_port = kwargs.pop('base_port', None)

    aligns = align_data_set(train_set)
    for eval_set in eval_sets:
        aligns += align_data_set(eval_set)

    aligned_iter = iter(ExecutableTuple(aligns).execute(session))
    datas, labels, sample_weights, init_scores = [], [], [], []
    for dataset in [train_set] + eval_sets:
        train_kw = dict()
        for arg in ['data', 'label', 'sample_weight', 'init_score']:
            if getattr(dataset, arg) is not None:
                train_kw[arg] = next(aligned_iter)
            else:
                train_kw[arg] = None

        datas.append(train_kw['data'])
        labels.append(train_kw['label'])
        sample_weights.append(train_kw['sample_weight'])
        init_scores.append(train_kw['init_score'])

    op = LGBMTrain(params=params, data=datas[0], label=labels[0], sample_weight=sample_weights[0],
                   init_score=init_scores[0], eval_datas=datas[1:], eval_labels=labels[1:],
                   eval_weights=sample_weights[1:], eval_init_score=init_scores[1:],
                   model_type=model_type, timeout=timeout, lgbm_port=base_port, kwds=kwargs)
    ret = op().execute(session=session, **run_kwargs).fetch(session=session)

    bst = pickle.loads(ret)
    evals_result.update(bst.evals_result_ or {})
    return bst



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import tensor as mt


# Use at least float64 for the accumulating functions to avoid precision issue
# see https://github.com/numpy/numpy/issues/9393. The float64 is also retained
# as it is in case the float overflows
def _safe_accumulator_op(op, x, *args, **kwargs):
    """
    This function provides numpy accumulator functions with a float64 dtype
    when used on a floating point input. This prevents accumulator overflow on
    smaller floating point dtypes.

    Parameters
    ----------
    op : function
        A accumulator function such as np.mean or np.sum
    x : numpy array
        A tensor to apply the accumulator function
    *args : positional arguments
        Positional arguments passed to the accumulator function after the
        input x
    **kwargs : keyword arguments
        Keyword arguments passed to the accumulator function

    Returns
    -------
    result : The output of the accumulator function passed to this function
    """
    if np.issubdtype(x.dtype, np.floating) and x.dtype.itemsize < 8:
        result = op(x, *args, **kwargs, dtype=np.float64)
    else:
        result = op(x, *args, **kwargs)
    return result


def row_norms(X, squared=False):
    """Row-wise (squared) Euclidean norm of X.

    Performs no input validation.

    Parameters
    ----------
    X : array_like
        The input tensor
    squared : bool, optional (default = False)
        If True, return squared norms.

    Returns
    -------
    array_like
        The row-wise (squared) Euclidean norm of X.
    """

    norms = (X * X).sum(axis=1)
    if not squared:
        norms = mt.sqrt(norms)
    return norms



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


cdef class ClusterInfo:
    def __init__(self, bint standalone, int n_process, str address=None,
                 str serialization='pickle', str advertise_address=None):
        if address and ':' not in address:
            raise ValueError('address must contain port')

        self.standalone = standalone
        self.n_process = n_process
        self.address = address
        self.serialization = serialization
        self.advertise_address = advertise_address

        if self.address:
            self.location, port = self.address.split(':', 1)
            self.port = int(port)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from .utils import WorkerActor


class CustomLogFetchActor(WorkerActor):
    def __init__(self):
        super().__init__()
        self._dispatch_ref = None

    def post_create(self):
        from .dispatcher import DispatchActor

        self._dispatch_ref = self.ctx.actor_ref(DispatchActor.default_uid())
        self._dispatch_ref.register_free_slot(self.uid, 'custom_log')

    def fetch_logs(self, log_paths, offsets, sizes):
        result = []
        for i, log_path in enumerate(log_paths):
            log_result = dict()

            offset = offsets[i]
            size = sizes[i]

            with self.ctx.fileobject(log_path, mode='r') as f:
                if offset < 0:
                    # process negative offset
                    offset = max(os.path.getsize(log_path) + offset, 0)

                if offset:
                    f.seek(offset)

                log_result['log'] = f.read(size)
                log_result['offset'] = f.tell()

            result.append(log_result)

        return result



# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from ...scheduler import ResourceActor
from ...scheduler.__main__ import SchedulerApplication
from ...scheduler.utils import SchedulerActor, SchedulerClusterInfoActor
from .core import K8SServiceMixin, ReadinessActor, K8SPodsIPWatcher

logger = logging.getLogger(__name__)


class WorkerWatcherActor(SchedulerActor):
    watcher_cls = K8SPodsIPWatcher
    _watcher_running = True

    def __init__(self):
        super().__init__()
        self._resource_ref = None

    def post_create(self):
        super().post_create()

        logger.debug('Actor %s running in process %d', self.uid, os.getpid())
        self._cluster_info_ref = self.ctx.actor_ref(SchedulerClusterInfoActor.default_uid())
        self.ref().watch_workers(_tell=True)

    def pre_destroy(self):
        super().pre_destroy()
        self._watcher_running = False

    def watch_workers(self):
        from kubernetes import client, config

        cls = type(self)

        if os.environ.get('KUBE_API_ADDRESS'):  # pragma: no cover
            k8s_config = client.Configuration()
            k8s_config.host = os.environ['KUBE_API_ADDRESS']
        else:
            k8s_config = config.load_incluster_config()

        watcher = self.watcher_cls(k8s_config, os.environ['MARS_K8S_POD_NAMESPACE'])

        for workers in watcher.watch_workers():  # pragma: no branch
            if not cls._watcher_running:  # pragma: no cover
                break

            if self._resource_ref is None:
                self.set_schedulers(self._cluster_info_ref.get_schedulers())
                self._resource_ref = self.get_actor_ref(ResourceActor.default_uid())

            if self._resource_ref:  # pragma: no branch
                self._resource_ref.mark_workers_alive(workers)


class K8SSchedulerApplication(K8SServiceMixin, SchedulerApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._readiness_ref = None
        self._worker_watcher_ref = None

    def start(self):
        self.write_pid_file()
        super().start()
        self._worker_watcher_ref = self.pool.create_actor(
            WorkerWatcherActor, uid=WorkerWatcherActor.default_uid())
        self._readiness_ref = self.pool.create_actor(ReadinessActor, uid=ReadinessActor.default_uid())

    def stop(self):
        self._readiness_ref.destroy()
        self._worker_watcher_ref.destroy()
        super().stop()


main = K8SSchedulerApplication()

if __name__ == '__main__':   # pragma: no branch
    main()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

import numpy as np
import pandas as pd

from ...errors import StorageFull
from ...serialize import dataserializer
from ...utils import calc_data_size, lazy_import
from .core import StorageHandler, DataStorageDevice, ObjectStorageMixin, \
    SpillableStorageMixin, wrap_promised, register_storage_handler_cls

cp = lazy_import('cupy', globals=globals(), rename='cp')
cudf = lazy_import('cudf', globals=globals())


class CudaHandler(StorageHandler, ObjectStorageMixin, SpillableStorageMixin):
    storage_type = DataStorageDevice.CUDA

    def __init__(self, storage_ctx, proc_id=None):
        StorageHandler.__init__(self, storage_ctx, proc_id=proc_id)
        self._cuda_store_ref_attr = None
        self._cuda_size_limit_attr = None

    @property
    def _cuda_store_ref(self):
        if self._cuda_store_ref_attr is None:
            self._cuda_store_ref_attr = self._storage_ctx.promise_ref(
                self._storage_ctx.manager_ref.get_process_holder(
                    self._proc_id, DataStorageDevice.CUDA))
        return self._cuda_store_ref_attr

    @property
    def _cuda_size_limit(self):
        if self._cuda_size_limit_attr is None:
            self._cuda_size_limit_attr = self._cuda_store_ref.get_size_limit()
        return self._cuda_size_limit_attr

    @staticmethod
    def _obj_to_cuda(o):
        from cupy.cuda.memory import OutOfMemoryError
        from numba.cuda.cudadrv.driver import CudaAPIError
        try:
            if isinstance(o, np.ndarray):
                return cp.asarray(o)
            elif isinstance(o, pd.DataFrame):
                return cudf.DataFrame.from_pandas(o)
            elif isinstance(o, pd.Series):
                return cudf.Series.from_pandas(o)
            return o
        except OutOfMemoryError:
            raise StorageFull
        except CudaAPIError as ex:
            if ex.code == 1:  # CUDA_ERROR_INVALID_VALUE
                raise StorageFull
            raise

    @staticmethod
    def _obj_to_mem(o):
        if cp and isinstance(o, cp.ndarray):
            o = cp.asnumpy(o)
        elif cudf and isinstance(o, (cudf.DataFrame, cudf.Series)):
            o = o.to_pandas()
        return o

    @wrap_promised
    def get_objects(self, session_id, data_keys, serialize=False, _promise=False):
        objs = self._cuda_store_ref.get_objects(session_id, data_keys)
        if serialize:
            objs = [dataserializer.serialize(self._obj_to_mem(o)) for o in objs]
        return objs

    @wrap_promised
    def put_objects(self, session_id, data_keys, objs, sizes=None, shapes=None,
                    serialize=False, pin_token=None, _promise=False):
        objs = [self._deserial(obj) if serialize else obj for obj in objs]
        sizes = sizes or [calc_data_size(obj) for obj in objs]
        shapes = shapes or [getattr(obj, 'shape', None) for obj in objs]

        obj = None
        succ_keys, succ_objs, succ_sizes, succ_shapes = [], [], [], []
        affected_keys = []
        request_size, capacity = 0, 0
        try:
            for idx, key, obj, size, shape in zip(itertools.count(0), data_keys, objs, sizes, shapes):
                try:
                    obj = objs[idx] = self._obj_to_cuda(obj)
                    succ_objs.append(obj)
                    succ_keys.append(key)
                    succ_shapes.append(shape)
                    succ_sizes.append(size)
                except StorageFull:
                    affected_keys.append(key)
                    request_size += size
                    capacity = self._cuda_size_limit

            self._cuda_store_ref.put_objects(
                session_id, succ_keys, succ_objs, succ_sizes, pin_token=pin_token)
            self.register_data(session_id, succ_keys, succ_sizes, succ_shapes)

            if affected_keys:
                raise StorageFull(request_size=request_size, capacity=capacity,
                                  affected_keys=affected_keys)
        finally:
            del obj
            objs[:] = []
            succ_objs[:] = []

    def load_from_object_io(self, session_id, data_keys, src_handler, pin_token=None):
        return self._batch_load_objects(
            session_id, data_keys,
            lambda keys: src_handler.get_objects(session_id, keys, _promise=True),
            pin_token=pin_token, batch_get=True)

    def load_from_bytes_io(self, session_id, data_keys, src_handler, pin_token=None):
        def _read_serialized(reader):
            with reader:
                return reader.get_io_pool().submit(reader.read).result()

        return self._batch_load_objects(
            session_id, data_keys,
            lambda k: src_handler.create_bytes_reader(session_id, k, _promise=True).then(_read_serialized),
            serialize=True, pin_token=pin_token,
        )

    def delete(self, session_id, data_keys, _tell=False):
        self._cuda_store_ref.delete_objects(session_id, data_keys, _tell=_tell)
        self.unregister_data(session_id, data_keys, _tell=_tell)

    def spill_size(self, size, multiplier=1):
        return self._cuda_store_ref.spill_size(size, multiplier, _promise=True)

    def lift_data_keys(self, session_id, data_keys):
        self._cuda_store_ref.lift_data_keys(session_id, data_keys, _tell=True)

    def pin_data_keys(self, session_id, data_keys, token):
        return self._cuda_store_ref.pin_data_keys(session_id, data_keys, token)

    def unpin_data_keys(self, session_id, data_keys, token, _tell=False):
        return self._cuda_store_ref.unpin_data_keys(session_id, data_keys, token, _tell=_tell)


register_storage_handler_cls(DataStorageDevice.CUDA, CudaHandler)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import functools
import io
import os
import shutil
import sys
import tempfile
import textwrap
import weakref

from .context import RunningMode, DistributedContext


_custom_log_dir = None


def _get_custom_log_dir():
    from .config import options

    global _custom_log_dir

    if _custom_log_dir is None:
        log_dir = options.custom_log_dir

        if log_dir is None:
            log_dir = tempfile.mkdtemp(prefix='mars-custom-log')

        _custom_log_dir = log_dir

        # remove log dir at exit
        atexit.register(lambda: shutil.rmtree(_custom_log_dir))

    return _custom_log_dir


class _LogWrapper:
    def __init__(self, ctx: DistributedContext, op,
                 log_path: str):
        self.ctx = ctx
        self.op = op
        self.log_path = log_path

        self.file = open(log_path, 'w')
        self.stdout = sys.stdout

        self.raw_stdout = self.stdout
        while isinstance(self.raw_stdout, _LogWrapper):
            self.raw_stdout = self.raw_stdout.stdout

        # flag about registering log path
        self.is_log_path_registered = False

    def __enter__(self):
        self.file.__enter__()
        # set stdout
        sys.stdout = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.__exit__(exc_type, exc_val, exc_tb)
        # set back stdout
        sys.stdout = self.stdout

    def _register_log_path(self):
        if self.is_log_path_registered:
            return

        # register log path
        session_id = self.ctx.session_id
        tileable_op_key = self.op.tileable_op_key
        chunk_op_key = self.op.key
        worker_addr = self.ctx.get_local_address()
        log_path = self.log_path

        custom_log_meta = self.ctx.get_custom_log_meta_ref()
        custom_log_meta.record_custom_log_path(
            session_id, tileable_op_key, chunk_op_key,
            worker_addr, log_path)

        self.is_log_path_registered = True

    def write(self, data):
        self._register_log_path()

        # write into file
        self.file.write(data)
        # force flush to make sure `fetch_log` can get stdout in time
        self.file.flush()
        # write into previous stdout
        self.raw_stdout.write(data)

    def flush(self):
        self.raw_stdout.flush()


def gen_log_path(session_id, op_key):
    filename = f"{str(session_id).replace('-', '_')}_{op_key}"
    custom_log_dir = _get_custom_log_dir()
    return os.path.join(custom_log_dir, filename)


def redirect_custom_log(func):
    """
    Redirect stdout to a file by wrapping ``Operand.execute(ctx, op)``
    """

    @functools.wraps(func)
    def wrap(cls, ctx: DistributedContext, op):
        # import inside, or Ray backend may fail
        from .config import options

        if getattr(ctx, 'running_mode', RunningMode.local) == RunningMode.local or \
                options.custom_log_dir is None:
            # do nothing for local scheduler
            return func(cls, ctx, op)

        log_path = gen_log_path(ctx.session_id, op.key)

        with _LogWrapper(ctx, op, log_path):
            return func(cls, ctx, op)

    return wrap


_tileable_to_log_fetcher = weakref.WeakKeyDictionary()


class LogFetcher:
    def __init__(self, tileable_op_key, session):
        self._tileable_op_key = tileable_op_key
        self._session = session
        self._chunk_op_key_to_result = dict()
        self._chunk_op_key_to_offsets = dict()

    def __len__(self):
        return len(self._chunk_op_key_to_result)

    @property
    def chunk_op_keys(self):
        return list(self._chunk_op_key_to_result.keys())

    @property
    def results(self):
        return list(self._chunk_op_key_to_result.values())

    @property
    def offsets(self):
        return list(self._chunk_op_key_to_offsets.values())

    def fetch(self, offsets=None, sizes=None):
        if offsets is None:
            offsets = self._chunk_op_key_to_offsets

        if sizes is None:
            sizes = 1 * 1024 ** 2  # 1M each time

        result: dict = self._session.fetch_tileable_op_logs(
            self._tileable_op_key, offsets=offsets, sizes=sizes)

        if result is None:
            return

        for chunk_key, chunk_result in result.items():
            self._chunk_op_key_to_result[chunk_key] = chunk_result['log']
            self._chunk_op_key_to_offsets[chunk_key] = chunk_result['offset']

    def _display(self, representation=True):
        if len(self) == 1:
            content = next(iter(self._chunk_op_key_to_result.values()))
            return repr(content) if representation else str(content)

        sio = io.StringIO()
        for chunk_op_key, content in self._chunk_op_key_to_result.items():
            sio.write(textwrap.dedent(
                f"""
                Chunk op key: {chunk_op_key}
                Out:
                {content}"""))
        result = sio.getvalue()
        return repr(result) if representation else str(result)

    def __repr__(self):
        return self._display(True)

    def __str__(self):
        return self._display(False)


def fetch(tileables, session, offsets=None, sizes=None):
    log_fetchers = []
    for tileable in tileables:
        tileable = tileable.data if hasattr(tileable, 'data') else tileable

        if tileable not in _tileable_to_log_fetcher:
            _tileable_to_log_fetcher[tileable] = LogFetcher(tileable.op.key, session)

        log_fetcher = _tileable_to_log_fetcher[tileable]
        log_fetcher.fetch(offsets=offsets, sizes=sizes)
        log_fetchers.append(log_fetcher)
    return log_fetchers



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyUnresolvedReferences
from ..core import GroupBy, DataFrameGroupBy, SeriesGroupBy


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, GROUPBY_TYPE, DATAFRAME_GROUPBY_TYPE
    from .core import groupby
    from .aggregation import agg
    from .apply import groupby_apply
    from .cum import cumcount, cummin, cummax, cumprod, cumsum
    from .getitem import df_groupby_getitem
    from .head import head
    from .transform import groupby_transform
    from .sample import groupby_sample

    for cls in DATAFRAME_TYPE:
        setattr(cls, 'groupby', groupby)

    for cls in SERIES_TYPE:
        setattr(cls, 'groupby', groupby)

    for cls in GROUPBY_TYPE:
        setattr(cls, 'agg', agg)
        setattr(cls, 'aggregate', agg)

        setattr(cls, 'sum', lambda groupby, **kw: agg(groupby, 'sum', **kw))
        setattr(cls, 'prod', lambda groupby, **kw: agg(groupby, 'prod', **kw))
        setattr(cls, 'max', lambda groupby, **kw: agg(groupby, 'max', **kw))
        setattr(cls, 'min', lambda groupby, **kw: agg(groupby, 'min', **kw))
        setattr(cls, 'count', lambda groupby, **kw: agg(groupby, 'count', **kw))
        setattr(cls, 'size', lambda groupby, **kw: agg(groupby, 'size', **kw))
        setattr(cls, 'mean', lambda groupby, **kw: agg(groupby, 'mean', **kw))
        setattr(cls, 'var', lambda groupby, **kw: agg(groupby, 'var', **kw))
        setattr(cls, 'std', lambda groupby, **kw: agg(groupby, 'std', **kw))
        setattr(cls, 'all', lambda groupby, **kw: agg(groupby, 'all', **kw))
        setattr(cls, 'any', lambda groupby, **kw: agg(groupby, 'any', **kw))
        setattr(cls, 'skew', lambda groupby, **kw: agg(groupby, 'skew', **kw))
        setattr(cls, 'kurt', lambda groupby, **kw: agg(groupby, 'kurt', **kw))
        setattr(cls, 'kurtosis', lambda groupby, **kw: agg(groupby, 'kurtosis', **kw))
        setattr(cls, 'sem', lambda groupby, **kw: agg(groupby, 'sem', **kw))

        setattr(cls, 'apply', groupby_apply)
        setattr(cls, 'transform', groupby_transform)

        setattr(cls, 'cumcount', cumcount)
        setattr(cls, 'cummin', cummin)
        setattr(cls, 'cummax', cummax)
        setattr(cls, 'cumprod', cumprod)
        setattr(cls, 'cumsum', cumsum)

        setattr(cls, 'head', head)

        setattr(cls, 'sample', groupby_sample)

    for cls in DATAFRAME_GROUPBY_TYPE:
        setattr(cls, '__getitem__', df_groupby_getitem)


_install()
del _install



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


cdef class ActorRef:
    cdef public str address
    cdef public object uid
    cdef public object _ctx

    cpdef object send(self, object message, bint wait=*, object callback=*)
    cpdef object tell(self, object message, object delay=*, bint wait=*,
                      object callback=*)
    cpdef object destroy(self, bint wait=*, object callback=*)


cdef class Actor:
    cdef str _address
    cdef object _uid
    cdef object _ctx

    cpdef ActorRef ref(self)
    cpdef post_create(self)
    cpdef on_receive(self, message)
    cpdef pre_destroy(self)


cdef class _FunctionActor(Actor):
    cpdef on_receive(self, message)


cpdef object create_actor_pool(str address=*, int n_process=*, object distributor=*,
                               object parallel=*, str backend=*, str advertise_address=*,
                               object pool_cls=*)
cpdef object new_client(object parallel=*, str backend=*)
